#!/usr/bin/python
# -*- coding: utf-8 -*-
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTIBILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

"Módulo Trazabilidad de Productos Veterinarios SENASA Resolución 369/2013"

# Información adicional y documentación:
# http://www.sistemasagiles.com.ar/trac/wiki/TrazabilidadProductosVeterinarios

__author__ = "Mariano Reingart <reingart@gmail.com>"
__copyright__ = "Copyright (C) 2014 Mariano Reingart"
__license__ = "GPL 3.0+"
__version__ = "1.11d"

# http://senasa.servicios.pami.org.ar/

import os
import socket
import sys
import datetime, time
import traceback
import py3simplesoap.client
from py3simplesoap.client import SoapClient, SoapFault, parse_proxy, \
                                set_http_wrapper
from py3simplesoap.simplexml import SimpleXMLElement
from cStringIO import StringIO

# importo funciones compartidas:
from .utils import leer, escribir, leer_dbf, guardar_dbf, N, A, I, json, \
                  dar_nombre_campo_dbf, get_install_dir, BaseWS, \
                  inicializar_y_capturar_excepciones

HOMO = True
TYPELIB = False

WSDL = "https://servicios.pami.org.ar/trazaenvet.WebService?wsdl"
LOCATION = "https://servicios.pami.org.ar/trazaenvet.WebService?wsdl"

# Formato de TransaccionSenasaDTO (SaveTransaccion)
TRANSACCION_DTO = [
    ('gln_origen', 13, A),
    ('gln_destino', 13, A),
    ('f_operacion', 10, A),
    ('f_elaboracion', 10, A),
    ('f_vto', 10, A),
    ('id_evento', 15, N),
    ('cod_producto', 14, A),
    ('n_cantidad', 30, N),
    ('n_serie', 20, A),
    ('n_lote', 50, A),
    ('n_cai', 15, A),
    ('n_cae', 15, A),
    ('id_motivo_destruccion', 5, A),
    ('n_manifiesto', 15, N),
    ('en_transporte', 1, A), # boolean
    ('n_remito', 15, A),
    ('motivo_devolucion', 100, A),
    ('observaciones', 1000, A),
    ('n_vale_compra', 15, A),
    ('apellidoNombres', 255, A),
    ('direccion', 200, A),
    ('numero', 6, N),
    ('localidad', 15, A),
    ('provincia', 15, A),
    ('n_postal', 8, A),
    ('cuit', 11, A),
    ('codigo_transaccion', 14, A),
]

# Formato para TransaccionSenasa (getTransacciones)
TRANSACCIONES = [
    ('id_transaccion_global', 15, N),
    ('id_transaccion', 15, N),
    ('f_transaccion', 10, A),
    ('f_operacion', 10, A),
    ('f_vencimiento', 10, A),
    ('f_elaboracion', 10, A),
    ('d_evento', 100, A),
    ('n_cantidad', 30, N),
    ('id_unidad', 15, N),
    ('d_unidad', 100, A),
    ('cod_producto', 14, A),
    ('id_unidad', 15, N),
    ('n_serie', 20, A),
    ('n_lote', 50, A),
    ('n_cai', 15, A),
    ('n_cae', 15, A),
    ('d_motivo_destruccion', 50, A),
    ('d_manifiesto', 15, A),
    ('en_transporte', 1, A),
    ('n_remito', 30, A),
    ('motivo_devolucion', 200, A),
    ('observaciones', 1000, A),
    ('n_vale_compra', 15, A),
    ('apellidoNombres', 255, A),
    ('direccion', 200, A),
    ('numero', 6, A),
    ('localidad', 250, A),
    ('provincia', 250, A),
    ('n_postal', 8, A),
    ('cuit', 11, A),
    ('d_agente_informador', 255, A),
    ('d_agente_origen', 255, A),
    ('d_agente_destino', 255, A),
    ('d_producto', 250, A),
    ('d_estado_transaccion', 30, A),
    ('d_tipo_transaccion', 30, A),
]

# Formato para Errores
ERRORES = [
    ('_c_error', 4, A),                 # código
    ('_d_error', 250, A),               # descripción
    ]


class TrazaVet(BaseWS):
    "Interfaz para el WebService de Trazabilidad de Veterinarios SENASA"

    _public_methods_ = ['SaveTransaccion', 'SendCancelaTransac',
                        'SendConfirmaTransacc', 'SendAlertaTransacc',
                        'GetTransacciones',
                        'Conectar', 'LeerError', 'LeerTransaccion',
                        'SetUsername',
                        'SetParametro', 'GetParametro',
                        'GetCodigoTransaccion', 'GetResultado', 'LoadTestXML']

    _public_attrs_ = [
        'Username', 'Password',
        'CodigoTransaccion', 'Errores', 'Resultado',
        'XmlRequest', 'XmlResponse',
        'Version', 'InstallDir',
        'Traceback', 'Excepcion',
        'CantPaginas', 'HayError', 'TransaccionSenasa',
        ]

    _reg_progid_ = "TrazaVet"
    _reg_clsid_ = "{2E28D79D-23B0-4C1E-85B9-64BDC41B8FCF}"

    # Variables globales para BaseWS:
    HOMO = HOMO
    WSDL = WSDL
    Version = "%s %s %s" % (__version__, HOMO and 'Homologación' or '',
                            py3simplesoap.client.__version__)

    def __init__(self, reintentos=1):
        self.Username = self.Password = None
        self.TransaccionSenasa = []
        BaseWS.__init__(self, reintentos)

    def inicializar(self):
        BaseWS.inicializar(self)
        self.CodigoTransaccion = self.Errores = self.Resultado = None
        self.Resultado = ''
        self.Errores = []   # lista de strings para la interfaz
        self.errores = []   # lista de diccionarios (uso interno)
        self.CantPaginas = self.HayError = None

    def __analizar_errores(self, ret):
        "Comprueba y extrae errores si existen en la respuesta XML"
        self.errores = ret.get('errores', [])
        self.Errores = ["%s: %s" % (it['c_error'], it['d_error'])
                        for it in ret.get('errores', [])]
        self.Resultado = ret.get('resultado')

    def Conectar(self, cache=None, wsdl=None, proxy="", wrapper=None, cacert=None, timeout=30):
        # Conecto usando el método estandard:
        ok = BaseWS.Conectar(self, cache, wsdl, proxy, wrapper, cacert, timeout,
                              soap_server="jetty")
        if ok:
            # si el archivo es local, asumo que ya esta corregido:
            if not self.wsdl.startswith("file"):
                # corrijo ubicación del servidor (localhost:59050 en el WSDL)
                location = self.wsdl[:-5]
                ws = self.client.services['IWebServiceSenasa']
                ws['ports']['IWebServiceSenasaPort']['location'] = location

            # Establecer credenciales de seguridad:
            self.client['wsse:Security'] = {
                'wsse:UsernameToken': {
                    'wsse:Username': self.Username,
                    'wsse:Password': self.Password,
                    }
                }
        return ok

    @inicializar_y_capturar_excepciones
    def SaveTransaccion(self, usuario, password,
                        gln_origen=None, gln_destino=None,
                        f_operacion=None, f_elaboracion=None, f_vto=None,
                        id_evento=None, cod_producto=None, n_cantidad=None,
                        n_serie=None, n_lote=None, n_cai=None, n_cae=None,
                        id_motivo_destruccion=None, n_manifiesto=None,
                        en_transporte=None, n_remito=None,
                        motivo_devolucion=None, observaciones=None,
                        n_vale_compra=None, apellidoNombres=None,
                        direccion=None, numero=None, localidad=None,
                        provincia=None, n_postal=None, cuit=None
                         ):
        "Realiza el registro de una transacción de productos fitosanitarios. "
        # creo los parámetros para esta llamada
        params = {  'gln_origen': gln_origen,
                    'gln_destino': gln_destino,
                    'f_operacion': f_operacion,
                    'f_elaboracion': f_elaboracion,
                    'f_vto': f_vto,
                    'id_evento': id_evento,
                    'cod_producto': cod_producto,
                    'n_cantidad': n_cantidad,
                    'n_serie': n_serie,
                    'n_lote': n_lote,
                    'n_cai': n_cai or None,
                    'n_cae': n_cae or None,
                    'id_motivo_destruccion': id_motivo_destruccion or None,
                    'n_manifiesto': n_manifiesto or None,
                    'en_transporte': en_transporte or None,
                    'n_remito': n_remito or None,
                    'motivo_devolucion': motivo_devolucion or None,
                    'observaciones': observaciones or None,
                    'n_vale_compra': n_vale_compra or None,
                    'apellidoNombres': apellidoNombres or None,
                    'direccion': direccion or None,
                    'numero': numero or None,
                    'localidad': localidad or None,
                    'provincia': provincia or None,
                    'n_postal': n_postal or None,
                    'cuit': cuit or None,
                    }
        res = self.client.saveTransacciones(
            arg0=params,
            arg1=usuario,
            arg2=password,
        )
        ret = res['return']
        self.CodigoTransaccion = ret.get('codigoTransaccion')
        self.__analizar_errores(ret)
        return True

    @inicializar_y_capturar_excepciones
    def SendCancelaTransac(self, usuario, password, codigo_transaccion):
        " Realiza la cancelación de una transacción"
        res = self.client.sendCancelaTransac(
            arg0=codigo_transaccion,
            arg1=usuario,
            arg2=password,
        )
        ret = res['return']
        self.CodigoTransaccion = ret.get('codigoTransaccion')
        self.__analizar_errores(ret)
        return True

    @inicializar_y_capturar_excepciones
    def SendConfirmaTransacc(self, usuario, password, p_ids_transac, f_operacion, n_cantidad=None):
        "Confirma la recepción de un medicamento"
        res = self.client.sendConfirmaTransacc(
            arg0=usuario,
            arg1=password,
            arg2={'p_ids_transac': p_ids_transac, 'f_operacion': f_operacion,
                  'n_cantidad': n_cantidad,
                 },
        )
        ret = res['return']
        self.CodigoTransaccion = ret.get('codigoTransaccion')
        self.__analizar_errores(ret)
        return True

    @inicializar_y_capturar_excepciones
    def SendAlertaTransacc(self, usuario, password, p_ids_transac_ws):
        "Alerta un medicamento, acción contraria a “confirmar la transacción”."
        res = self.client.sendAlertaTransacc(
            arg0=usuario,
            arg1=password,
            arg2=p_ids_transac_ws,
        )
        ret = res['return']
        self.CodigoTransaccion = ret.get('id_transac_asociada')
        self.__analizar_errores(ret)
        return True

    @inicializar_y_capturar_excepciones
    def GetTransacciones(self, usuario, password,
                id_transaccion=None, id_evento=None, gln_origen=None,
                fecha_desde_t=None, fecha_hasta_t=None,
                fecha_desde_v=None, fecha_hasta_v=None,
                gln_informador=None, id_tipo_transaccion=None,
                gtin_elemento=None, n_lote=None, n_serie=None,
                n_remito_factura=None,
                ):
        "Trae un listado de las transacciones que no están confirmadas"

        # preparo los parametros de entrada opcionales:
        kwargs = {}
        if id_transaccion is not None:
            kwargs['arg2'] = id_transaccion
        if id_evento is not None:
            kwargs['arg3'] = id_evento
        if gln_origen is not None:
            kwargs['arg4'] = gln_origen
        if fecha_desde_t is not None:
            kwargs['arg5'] = fecha_desde_t
        if fecha_hasta_t is not None:
            kwargs['arg6'] = fecha_hasta_t
        if fecha_desde_v is not None:
            kwargs['arg7'] = fecha_desde_v
        if fecha_hasta_v is not None:
            kwargs['arg8'] = fecha_hasta_v
        if gln_informador is not None:
            kwargs['arg9'] = gln_informador
        if id_tipo_transaccion is not None:
            kwargs['arg10'] = id_tipo_transaccion
        if gtin_elemento is not None:
            kwargs['arg11'] = gtin_elemento
        if n_lote is not None:
            kwargs['arg12'] = n_lote
        if n_serie is not None:
            kwargs['arg13'] = n_serie
        if n_remito_factura is not None:
            kwargs['arg14'] = n_remito_factura

        # llamo al webservice
        res = self.client.getTransacciones(
            arg0=usuario,
            arg1=password,
            **kwargs
        )
        ret = res['return']
        if ret:
            self.__analizar_errores(ret)
            self.CantPaginas = ret.get('cantPaginas')
            self.HayError = ret.get('hay_error')
            self.TransaccionSenasa = [it for it in ret.get('list', [])]
        return True

    def  LeerTransaccion(self):
        "Recorro TransaccionSenasa devuelto por GetTransacciones"
         # usar GetParametro para consultar el valor retornado por el webservice

        if self.TransaccionSenasa:
            # extraigo el primer item
            self.params_out = self.TransaccionSenasa.pop(0)
            return True
        else:
            # limpio los parámetros
            self.params_out = {}
            return False

    def LeerError(self):
        "Recorro los errores devueltos y devuelvo el primero si existe"

        if self.Errores:
            # extraigo el primer item
            er = self.Errores.pop(0)
            return er
        else:
            return ""

    def SetUsername(self, username):
        "Establezco el nombre de usuario"
        self.Username = username

    def SetPassword(self, password):
        "Establezco la contraseña"
        self.Password = password

    def GetCodigoTransaccion(self):
        "Devuelvo el código de transacción"
        return self.CodigoTransaccion

    def GetResultado(self):
        "Devuelvo el resultado"
        return self.Resultado


def main():
    "Función principal de pruebas (obtener CAE)"
    import os, time, sys
    global WSDL, LOCATION

    DEBUG = '--debug' in sys.argv

    ws = TrazaVet()

    ws.Username = 'testwservice'
    ws.Password = 'testwservicepsw'

    if '--prod' in sys.argv and not HOMO:
        WSDL = "https://servicios.pami.org.ar/trazavet.WebService?wsdl"
        print("Usando WSDL:", WSDL)
        sys.argv.pop(sys.argv.index("--prod"))

    if '--proxy' in sys.argv and not HOMO:
        proxy = sys.argv.pop(sys.argv.index("--proxy") + 1)
        print("Usando proxy:", proxy)
        sys.argv.pop(sys.argv.index("--proxy"))
    else:
        proxy = None

    # Inicializo las variables y estructuras para el archivo de intercambio:
    transaccion_dto = []
    transacciones = []
    errores = []
    formatos = [('TransaccionDTO', TRANSACCION_DTO, transaccion_dto),
                ('Transacciones', TRANSACCIONES, transacciones),
                ('Errores', ERRORES, errores),
               ]

    if '--formato' in sys.argv:
        print("Formato:")
        for msg, formato, lista in formatos:
            comienzo = 1
            print("=== %s ===" % msg)
            print("|| %-25s || %-12s || %-5s || %-4s || %-10s ||" % (
                "Nombre", "Tipo", "Long.", "Pos(txt)", "Campo(dbf)"))
            claves = []
            for fmt in formato:
                clave, longitud, tipo = fmt[0:3]
                clave_dbf = dar_nombre_campo_dbf(clave, claves)
                claves.append(clave_dbf)
                print("|| %-25s || %-12s || %5d ||   %4d   || %-10s ||" % (
                    clave, tipo, longitud, comienzo, clave_dbf))
                comienzo += longitud
        sys.exit(0)

    if '--cargar' in sys.argv:
        if '--dbf' in sys.argv:
            leer_dbf(formatos[:1], {})
        elif '--json' in sys.argv:
            for formato in formatos[:1]:
                archivo = open(formato[0].lower() + ".json", "r")
                d = json.load(archivo)
                formato[2].extend(d)
                archivo.close()
        else:
            for formato in formatos[:1]:
                archivo = open(formato[0].lower() + ".txt", "r")
                for linea in archivo:
                    d = leer(linea, formato[1])
                    formato[2].append(d)
                archivo.close()

    ws.Conectar("", WSDL, proxy)

    if ws.Excepcion:
        print(ws.Excepcion)
        print(ws.Traceback)
        sys.exit(-1)

    # Datos de pruebas:

    if '--test' in sys.argv:
        transaccion_dto.append(dict(
            gln_origen="9876543210982", gln_destino="3692581473693",
            f_operacion=datetime.datetime.now().strftime("%d/%m/%Y"),
            f_elaboracion=datetime.datetime.now().strftime("%d/%m/%Y"),
            f_vto=(datetime.datetime.now()+datetime.timedelta(30)).strftime("%d/%m/%Y"),
            id_evento=11,
            cod_producto="88900000000001",
            n_cantidad=1,
            n_serie=int(time.time()*10),
            n_lote=datetime.datetime.now().strftime("%Y"),
            n_cai="123456789012345",
            n_cae="",
            id_motivo_destruccion=0,
            n_manifiesto="",
            en_transporte="N",
            n_remito="1234",
            motivo_devolucion="",
            observaciones="prueba",
            n_vale_compra="",
            apellidoNombres="Juan Peres",
            direccion="Saraza", numero="1234",
            localidad="Hurlingham", provincia="Buenos Aires",
            n_postal="1688",
            cuit="20267565393",
            codigo_transaccion=None,
        ))

    # Opciones principales:

    if '--confirma' in sys.argv:
        if '--loadxml' in sys.argv:
            ws.LoadTestXML("trazamed_confirma.xml")  # cargo respuesta
            ok = ws.SendConfirmaTransacc(usuario="pruebasws", password="pruebasws",
                                   p_ids_transac="1", f_operacion="31-12-2013")
            if not ok:
                raise RuntimeError(ws.Excepcion)
        ws.SendConfirmaTransacc(*sys.argv[sys.argv.index("--confirma")+1:])
    elif '--alerta' in sys.argv:
        ws.SendAlertaTransacc(*sys.argv[sys.argv.index("--alerta")+1:])
    elif '--cancela' in sys.argv:
        ws.SendCancelaTransac(*sys.argv[sys.argv.index("--cancela")+1:])
    elif '--consulta' in sys.argv:
        ws.GetTransacciones(
                            *sys.argv[sys.argv.index("--consulta")+1:]
                            )
        print("CantPaginas", ws.CantPaginas)
        print("HayError", ws.HayError)
        # print("TransaccionSenasa", ws.TransaccionSenasa)
        # parametros comunes de salida (columnas de la tabla):
        claves = [k for k, v, l in TRANSACCIONES]
        # extiendo la lista de resultado para el archivo de intercambio:
        transacciones.extend(ws.TransaccionSenasa)
        # encabezado de la tabla:
        print("||", "||".join(["%s" % clave for clave in claves]), "||")
        # recorro los datos devueltos (TransaccionSenasa):
        while ws.LeerTransaccion():
            for clave in claves:
                print("||", ws.GetParametro(clave))         # imprimo cada fila
            print("||")
    else:
        argv = [argv for argv in sys.argv if not argv.startswith("--")]
        if not transaccion_dto:
            if len(argv)>10:
                ws.SaveTransaccion(*argv[1:])
            else:
                print("ERROR: no se indicaron todos los parámetros requeridos")
        elif transaccion_dto:
            try:
                usuario, password = argv[-2:]
            except:
                print("ADVERTENCIA: no se indico parámetros usuario y passoword")
                usuario, password = "senasaws", "Clave2013"
            for i, dto in enumerate(transaccion_dto):
                print("Procesando registro", i)
                del dto['codigo_transaccion']
                ws.SaveTransaccion(usuario, password, **dto)
                dto['codigo_transaccion'] = ws.CodigoTransaccion
                errores.extend(ws.errores)
                print("|Resultado %5s|CodigoTransaccion %10s|Errores|%s|" % (
                    ws.Resultado,
                    ws.CodigoTransaccion,
                    '|'.join(ws.Errores or []),
                    ))
        else:
            print("ERROR: no se especificaron productos a informar")

    if not transaccion_dto:
        print("|Resultado %5s|CodigoTransaccion %10s|Errores|%s|" % (
                ws.Resultado,
                ws.CodigoTransaccion,
                '|'.join(ws.Errores or []),
                ))

    if ws.Excepcion:
        print(ws.Traceback)

    if '--grabar' in sys.argv:
        if '--dbf' in sys.argv:
            guardar_dbf(formatos, True, {})
        elif '--json' in sys.argv:
            for formato in formatos:
                archivo = open(formato[0].lower() + ".json", "w")
                json.dump(formato[2], archivo, sort_keys=True, indent=4)
                archivo.close()
        else:
            for formato in formatos:
                archivo = open(formato[0].lower() + ".txt", "w")
                for it in formato[2]:
                    archivo.write(escribir(it, formato[1]))
            archivo.close()


# busco el directorio de instalación (global para que no cambie si usan otra dll)
INSTALL_DIR = TrazaVet.InstallDir = get_install_dir()


if __name__ == '__main__':

    # ajusto el encoding por defecto (si se redirije la salida)
    if not hasattr(sys.stdout, "encoding") or sys.stdout.encoding is None:
        import codecs, locale
        sys.stdout = codecs.getwriter(locale.getpreferredencoding())(sys.stdout,"replace");
        sys.stderr = codecs.getwriter(locale.getpreferredencoding())(sys.stderr,"replace");

    if '--register' in sys.argv or '--unregister' in sys.argv:
        import pythoncom
        import win32com.server.register
        win32com.server.register.UseCommandLine(TrazaVet)
    elif "/Automate" in sys.argv:
        # MS seems to like /automate to run the class factories.
        import win32com.server.localserver
        #win32com.server.localserver.main()
        # start the server.
        win32com.server.localserver.serve([TrazaVet._reg_clsid_])
    else:
        main()



#!/usr/bin/python
# -*- coding: utf-8 -*-
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation; either version 3, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTIBILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

"Implementación pythónica de cliente SOAP"

__author__ = "Mariano Reingart (mariano@nsis.com.ar)"
__copyright__ = "Copyright (C) 2008 Mariano Reingart"
__license__ = "LGPL 3.0"
__version__ = "1.0"

import httplib2
from py3simplesoap.simplexml import SimpleXMLElement

class SoapFault(RuntimeError):
    def __init__(self,faultcode,faultstring):
        self.faultcode = faultcode
        self.faultstring = faultstring

# soap protocol specification & namespace
soap_namespaces = dict(
    soap11="http://schemas.xmlsoap.org/soap/envelope/",
    soap="http://schemas.xmlsoap.org/soap/envelope/",
    soapenv="http://schemas.xmlsoap.org/soap/envelope/",
    soap12="http://www.w3.org/2003/05/soap-env",
)

class SoapClient(object):
    "Manejo de Cliente SOAP Simple (símil PHP)"
    def __init__(self, location = None, action = None, namespace = None,
                 cert = None, trace = False, exceptions = False, proxy = None, ns=False,
                 soap_ns=None):
        self.certssl = cert
        self.keyssl = None
        self.location = location        # server location (url)
        self.action = action            # SOAP base action
        self.namespace = namespace      # message
        self.trace = trace              # show debug messages
        self.exceptions = exceptions    # lanzar execpiones? (Soap Faults)
        self.xml_request = self.xml_response = ''
        if not soap_ns and not ns:
            self.__soap_ns = 'soap' # 1.1
        elif not soap_ns and ns:
            self.__soap_ns = 'soapenv' # 1.2
        else:
            self.__soap_ns = soap_ns

        if not proxy:
            self.http = httplib2.Http(timeout=60)
        else:
            import socks
            ##httplib2.debuglevel=4
            self.http = httplib2.Http(proxy_info = httplib2.ProxyInfo(
                proxy_type=socks.PROXY_TYPE_HTTP, **proxy))
        #if self.certssl: # esto funciona para validar al server?
        #    self.http.add_certificate(self.keyssl, self.keyssl, self.certssl)
        self.__ns = ns
        if not ns:
            self.__xml = """<?xml version="1.0" encoding="UTF-8"?>
<%(soap_ns)s:Envelope xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
    xmlns:xsd="http://www.w3.org/2001/XMLSchema"
    xmlns:%(soap_ns)s="%(soap_uri)s">
<%(soap_ns)s:Body>
    <%(method)s xmlns="%(namespace)s">
    </%(method)s>
</%(soap_ns)s:Body>
</%(soap_ns)s:Envelope>"""
        else:
            self.__xml = """<?xml version="1.0" encoding="UTF-8"?>
<%(soap_ns)s:Envelope xmlns:%(soap_ns)s="%(soap_uri)s" xmlns:%(ns)s="%(namespace)s">
<%(soap_ns)s:Header/>
<%(soap_ns)s:Body>
    <%(ns)s:%(method)s>
    </%(ns)s:%(method)s>
</%(soap_ns)s:Body>
</%(soap_ns)s:Envelope>"""

    def __getattr__(self, attr):
        "Devuelve un pseudo-método que puede ser llamado"
        return lambda self=self, *args, **kwargs: self.call(attr,*args,**kwargs)

    def call(self, method, *args, **kwargs):
        "Prepara el xml y realiza la llamada SOAP, devuelve un SimpleXMLElement"
        # Mensaje de Solicitud SOAP básico:
        xml = self.__xml % dict(method=method, namespace=self.namespace, ns=self.__ns,
                                soap_ns=self.__soap_ns, soap_uri=soap_namespaces[self.__soap_ns])
        request = SimpleXMLElement(xml,namespace=self.__ns and self.namespace, prefix=self.__ns)
        # parsear argumentos
        if kwargs:
            parameters = list(kwargs.items())
        else:
            parameters = args
        for k,v in parameters: # dict: tag=valor
            self.parse(getattr(request,method),k,v)
        self.xml_request = request.as_xml()
        self.xml_response = self.send(method, self.xml_request)
        response = SimpleXMLElement(self.xml_response, namespace=self.namespace)
        if self.exceptions and ("soapenv:Fault" in response or "soap:Fault" in response):
            raise SoapFault(str(response.faultcode), str(response.faultstring))
        return response

    def parse(self, node, tag, value, add_child=True):
        "Analiza un objeto y devuelve su representación XML"
        ns = self.__soap_ns!='soapenv' # not add ns to childs in for soap1.1
        if isinstance(value, dict):  # serializar diccionario (<key>value</key>)
            child = add_child and node.add_child(tag,ns=ns) or node
            for k,v in list(value.items()):
                self.parse(child, k, v)
        elif isinstance(value, tuple):  # serializar tupla(<key>value</key>)
            child = add_child and node.add_child(tag,ns=ns) or node
            for k,v in value:
                self.parse(getattr(node,tag), k, v)
        elif isinstance(value, list): # serializar listas
            child=node.add_child(tag,ns=ns)
            for t in value:
                self.parse(child,tag,t, False)
        elif isinstance(value, str): # no volver a convertir los strings y unicodes
            node.add_child(tag,value,ns=ns)
        else: # el resto de los objetos se convierten a string
            if value is not None:
                node.add_child(tag,str(value),ns=ns) # habria que agregar un método asXML?

    def send(self, method, xml):
        "Envía el pedido SOAP por HTTP (llama al método con el xml como cuerpo)"
        if self.location == 'test': return
        location = "%s" % self.location #?op=%s" % (self.location, method)
        headers={
                'Content-type': 'text/xml; charset="UTF-8"',
                'Content-length': str(len(xml)),
                "SOAPAction": "\"%s%s\"" % (self.action,method)
                }
        if self.trace:
            print("-"*80)
            print("POST %s" % location)
            print('\n'.join(["%s: %s" % (k,v) for k,v in list(headers.items())]))
            print("\n%s" % xml.decode("utf8","ignore"))
        response, content = self.http.request(
            location,"POST", body=xml, headers=headers )
        self.response = response
        self.content = content
        if self.trace:
            print()
            print('\n'.join(["%s: %s" % (k,v) for k,v in list(response.items())]))
            print(content.decode("utf8","ignore"))
            print("="*80)
        return content


def parse_proxy(proxy_str):
    "Parses proxy address user:pass@host:port into a dict suitable for httplib2"
    proxy_dict = {}
    if proxy_str is None:
        return
    if "@" in proxy_str:
        user_pass, host_port = proxy_str.split("@")
    else:
        user_pass, host_port = "", proxy_str
    if ":" in host_port:
        host, port = host_port.split(":")
        proxy_dict['proxy_host'], proxy_dict['proxy_port'] = host, int(port)
    if ":" in user_pass:
        proxy_dict['proxy_user'], proxy_dict['proxy_pass'] = user_pass.split(":")
    return proxy_dict


if __name__=="__main__":
    client = SoapClient(
        location = "https://fwshomo.afip.gov.ar/wsctg/services/CTGService",
        action = 'http://impl.service.wsctg.afip.gov.ar/CTGService/', # SOAPAction
        namespace = "http://impl.service.wsctg.afip.gov.ar/CTGService/",
        trace = True,
        ns = True)

    response = client.dummy()
    result = response.dummyResponse
    print(str(result.appserver))
    print(str(result.dbserver))
    print(str(result.authserver))

    sys.exit(0)


    # Demo & Test: Feriados (Ministerio del Interior):
    from datetime import datetime, timedelta
    client = SoapClient(
        location = "http://webservices.mininterior.gov.ar/Feriados/Service.svc",
        action = 'http://tempuri.org/IMyService/', # SOAPAction
        namespace = "http://tempuri.org/FeriadoDS.xsd",
        trace = True)
    dt1 = datetime.today() - timedelta(days=60)
    dt2 = datetime.today() + timedelta(days=60)
    feriadosXML = client.FeriadosEntreFechasAsXml(dt1=dt1.isoformat(), dt2=dt2.isoformat());
    print(feriadosXML)

    # print(parse_proxy(None))
    # print(parse_proxy("host:1234"))
    # print(parse_proxy("user:pass@host:1234"))
    # sys.exit(0)
    # Demo & Test:
    token = "PD94bWwgdmVyc2lvbj0iMS4wIiBlbmNvZGluZz0iVVRGLTgiIHN0YW5kYW"
    sign = "gXVvzVwRrfkUAZKoy8ZqA3AL8IZgVxUvOHQH6g1/XzZJns1/k0lUdJslkzW"
    cuit = int(30199999)
    id = 1234
    cbte =199
    client = SoapClient(
        location = "https://wswhomo.afip.gov.ar/wsfe/service.asmx",
        action = 'http://ar.gov.afip.dif.facturaelectronica/', # SOAPAction
        namespace = "http://ar.gov.afip.dif.facturaelectronica/",
        trace = True)
    results = client.FERecuperaQTYRequest(
        argAuth= {"Token": token, "Sign": sign, "cuit":int(cuit)}
    )
    if int(results.FERecuperaQTYRequestResult.RError.percode) != 0:
        print("Percode: %s" % results.FERecuperaQTYRequestResult.RError.percode)
        print("MSGerror: %s" % results.FERecuperaQTYRequestResult.RError.perrmsg)
    else:
        print(int(results.FERecuperaQTYRequestResult.qty.value))



#!/usr/bin/python
# -*- coding: utf-8 -*-
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTIBILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

"Módulo de Intefase para archivos de texto (mercado interno versión 1)"

__author__ = "Mariano Reingart (reingart@gmail.com)"
__copyright__ = "Copyright (C) 2010-2015 Mariano Reingart"
__license__ = "GPL 3.0"
__version__ = "1.36a"

import datetime
import os
import sys
import time
import traceback
import warnings
from ConfigParser import SafeConfigParser

# revisar la instalación de pyafip.ws:
import wsfev1
from .php import SimpleXMLElement, SoapClient, SoapFault, date
from .utils import leer, escribir, leer_dbf, guardar_dbf, N, A, I


HOMO = wsfev1.HOMO
DEBUG = False
XML = False
CONFIG_FILE = "rece.ini"

LICENCIA = """
rece1.py: Interfaz de texto para generar Facturas Electrónica Mercado Interno
V1
Copyright (C) 2010 Mariano Reingart reingart@gmail.com

Este progarma es software libre, se entrega ABSOLUTAMENTE SIN GARANTIA
y es bienvenido a redistribuirlo bajo la licencia GPLv3.

Para información adicional sobre garantía, soporte técnico comercial
e incorporación/distribución en programas propietarios ver PyAfipWs:
http://www.sistemasagiles.com.ar/trac/wiki/PyAfipWs
"""

# definición del formato del archivo de intercambio:

ENCABEZADO = [
    ('tipo_reg', 1, N),  # 0: encabezado
    ('fecha_cbte', 8, A),
    ('tipo_cbte', 2, N), ('punto_vta', 4, N),
    ('cbt_desde', 8, N),
    ('cbt_hasta', 8, N),
    ('concepto', 1, N),  # 1:bienes, 2:servicios,...
    ('tipo_doc', 2, N),  # 80
    ('nro_doc', 11, N),  # 50000000016
    ('imp_total', 15, I, 2),
    ('no_usar', 15, I, 2),
    ('imp_tot_conc', 15, I, 2),
    ('imp_neto', 15, I, 2),
    ('imp_iva', 15, I, 2),
    ('imp_trib', 15, I, 2),
    ('imp_op_ex', 15, I, 2),
    ('moneda_id', 3, A),
    ('moneda_ctz', 10, I, 6),  # 10,6
    ('fecha_venc_pago', 8, A),   # opcional solo conceptos 2 y 3
    ('cae', 14, A), ('fch_venc_cae', 8, A),
    ('resultado', 1, A),
    ('motivos_obs', 1000, A),
    ('err_code', 6, A),
    ('err_msg', 1000, A),
    ('reproceso', 1, A),
    ('emision_tipo', 4, A),
    ('fecha_serv_desde', 8, A),  # opcional solo conceptos 2 y 3
    ('fecha_serv_hasta', 8, A),  # opcional solo conceptos 2 y 3
    ]

# DETALLE = [
#     ('tipo_reg', 1, N), # 1: detalle item
#     ('codigo', 30, A),
#     ('qty', 12, I),
#     ('umed', 2, N),
#     ('precio', 12, I, 3),
#     ('imp_total', 14, I, 3),
#     ('ds', 4000, A),
#     ]

TRIBUTO = [
    ('tipo_reg', 1, N),  # 1: tributo
    ('tributo_id', 16, N),
    ('desc', 100, A),
    ('base_imp', 15, I, 2),
    ('alic', 15, I, 2),
    ('importe', 15, I, 2),
]

IVA = [
    ('tipo_reg', 1, N),  # 2: alicuota de IVA
    ('iva_id', 16, N),
    ('base_imp', 15, I, 2),
    ('importe', 15, I, 2),
]

CMP_ASOC = [
    ('tipo_reg', 1, N),  # 3: comprobante asociado
    ('tipo', 3, N), ('pto_vta', 4, N),
    ('nro', 8, N),
]

OPCIONAL = [
    ('tipo_reg', 1, N),  # 6: datos opcionales
    ('opcional_id', 4, A),
    ('valor', 250, A),
]

# Constantes (tablas de parámetros):

TIPO_CBTE = {
    1: "FAC A", 2: "N/D A", 3: "N/C A", 6: "FAC B", 7: "N/D B",
    8: "N/C B", 4: "REC A", 5: "NV A", 9: "REC B",
    10: "NV B", 11: "FAC C", 12: "N/D C", 13: "N/C C", 15: "REC C",
    49: "BIENES USADOS",
    60: "LIQ PROD A", 61: "LIQ PROD B", 63: "LIQ A", 64: "LIQ B",
}

TIPO_DOC = {80: 'CUIT', 86: 'CUIL', 96: 'DNI', 99: '', 87: u"CDI"}


def autorizar(ws, entrada, salida, informar_caea=False):
    encabezados = []
    if '/dbf' in sys.argv:
        tributos = []
        ivas = []
        cbtasocs = []
        encabezados = []
        opcionales = []
        if DEBUG:
            print("Leyendo DBF...")

        formatos = [('Encabezado', ENCABEZADO, encabezados),
                    ('Tributo', TRIBUTO, tributos),
                    ('Iva', IVA, ivas),
                    ('Comprobante Asociado', CMP_ASOC, cbtasocs),
                    ('Datos Opcionales', OPCIONAL, opcionales),
                    ]
        dic = leer_dbf(formatos, conf_dbf)

        # rearmar estructura asociando id (comparando, si se útiliza)
        for encabezado in encabezados:
            for tributo in tributos:
                if tributo.get("id") == encabezado.get("id"):
                    encabezado.setdefault("tributos", []).append(tributo)
            for iva in ivas:
                if iva.get("id") == encabezado.get("id"):
                    encabezado.setdefault("ivas", []).append(iva)
            for cbtasoc in cbtasocs:
                if cbtasoc.get("id") == encabezado.get("id"):
                    encabezado.setdefault("cbtasocs", []).append(cbtasoc)
            for opcional in opcionales:
                if opcional.get("id") == encabezado.get("id"):
                    encabezado.setdefault("opcionales", []).append(opcional)
            if encabezado.get("id") is None and len(encabezados) > 1:
                # compatibilidad hacia atrás, descartar si hay más de 1 factura
                warnings.warn("Para múltiples registros debe usar campo id!")
                break
    elif '/json' in sys.argv:
        # ya viene estructurado
        import json
        encabezados = json.load(entrada)
    else:
        # la estructura está implícita en el órden de los registros (líneas)
        for linea in entrada:
            if str(linea[0]) == '0':
                encabezado = leer(linea, ENCABEZADO)
                encabezados.append(encabezado)
                if DEBUG:
                    print(len(encabezados),
                          "Leida factura %(cbt_desde)s" % encabezado)
            elif str(linea[0]) == '1':
                tributo = leer(linea, TRIBUTO)
                encabezado.setdefault("tributos", []).append(tributo)
            elif str(linea[0]) == '2':
                iva = leer(linea, IVA)
                encabezado.setdefault("ivas", []).append(iva)
            elif str(linea[0]) == '3':
                cbtasoc = leer(linea, CMP_ASOC)
                encabezado.setdefault("cbtasocs", []).append(cbtasoc)
            elif str(linea[0]) == '6':
                opcional = leer(linea, OPCIONAL)
                encabezado.setdefault("opcionales", []).append(opcional)
            else:
                print("Tipo de registro incorrecto:", linea[0])

    if informar_caea:
        if '/testing' in sys.argv:
            encabezado['cae'] = '21073372218437'
        encabezado['caea'] = encabezado['cae']

    if not encabezados:
        raise RuntimeError("No se pudieron leer los registros de la entrada")

    # ajusto datos para pruebas en depuración (nro de cbte. / fecha)
    if '--testing' in sys.argv and DEBUG:
        encabezado['punto_vta'] = 9998
        cbte_nro = int(ws.CompUltimoAutorizado(encabezado['tipo_cbte'],
                                               encabezado['punto_vta'])) + 1
        encabezado['cbt_desde'] = cbte_nro
        encabezado['cbt_hasta'] = cbte_nro
        encabezado['fecha_cbte'] = datetime.datetime.now().strftime("%Y%m%d")

    # recorrer los registros para obtener CAE (dicts tendrá los procesados)
    dicts = []
    for encabezado in encabezados:
        # extraer sub-registros:
        ivas = encabezado.get('ivas', encabezado.get('iva', []))
        tributos = encabezado.get('tributos', [])
        cbtasocs = encabezado.get('cbtasocs', [])
        opcionales = encabezado.get('opcionales', [])

        ws.CrearFactura(**encabezado)
        for tributo in tributos:
            ws.AgregarTributo(**tributo)
        for iva in ivas:
            ws.AgregarIva(**iva)
        for cbtasoc in cbtasocs:
            ws.AgregarCmpAsoc(**cbtasoc)
        for opcional in opcionales:
            ws.AgregarOpcional(**opcional)

        if DEBUG:
            print('\n'.join(
                ["%s='%s'" % (k, str(v)) for k, v in ws.factura.items()]))
        if not DEBUG or raw_input("Facturar (S/n)?") == "S":
            if not informar_caea:
                cae = ws.CAESolicitar()
                dic = ws.factura
            else:
                cae = ws.CAEARegInformativo()
                dic = ws.factura
                print(
                    "Procesando %s %04d %08d %08d %s %s $ %0.2f IVA: $ \
                    %0.2f" % (
                        TIPO_CBTE.get(
                            dic['tipo_cbte'],
                            dic['tipo_cbte']
                        ),
                        dic['punto_vta'],
                        dic['cbt_desde'],
                        dic['cbt_hasta'],
                        TIPO_DOC.get(
                            dic['tipo_doc'],
                            dic['tipo_doc']),
                        dic['nro_doc'],
                        float(dic['imp_total']),
                        float(
                            dic['imp_iva'] if dic['imp_iva'] is not None
                            else 'NaN'
                        )
                    ))
            dic.update(encabezado)         # preservar la estructura leida
            dic.update({
                'cae': cae and str(cae) or '',
                'fch_venc_cae': ws.Vencimiento and str(ws.Vencimiento) or '',
                'resultado': ws.Resultado,
                'motivos_obs': ws.Obs,
                'err_code': str(ws.ErrCode),
                'err_msg': ws.ErrMsg,
                'reproceso': ws.Reproceso,
                'emision_tipo': ws.EmisionTipo,
                })
            dicts.append(dic)
            print(
                "NRO:", dic['cbt_desde'], "Resultado:", dic['resultado'],
                "%s:" % ws.EmisionTipo, dic['cae'], "Obs:",
                dic['motivos_obs'].encode("ascii", "ignore"), "Err:",
                dic['err_msg'].encode("ascii", "ignore"),
                "Reproceso:", dic['reproceso']
            )
    if dicts:
        escribir_facturas(dicts, salida)


def escribir_facturas(encabezados, archivo, agrega=False):
    if '/json' in sys.argv:
        import json
        facturas = []
        for dic in encabezados:
            factura = dic.copy()
            facturas.append(factura)
            # ajsutes por compatibilidad hacia atras y con pyfepdf
            factura['fecha_vto'] = factura.get('fch_venc_cae')
            if 'iva' in factura:
                factura['ivas'] = factura.get('iva', [])
                del factura['iva']
        json.dump(facturas, archivo, sort_keys=True, indent=4)
    else:
        for dic in encabezados:
            dic['tipo_reg'] = 0
            archivo.write(escribir(dic, ENCABEZADO))
            if 'tributos' in dic:
                for it in dic['tributos']:
                    it['tipo_reg'] = 1
                    archivo.write(escribir(it, TRIBUTO))
            if 'iva' in dic or 'ivas' in dic:
                for it in dic.get('iva', dic.get('ivas')):
                    it['tipo_reg'] = 2
                    archivo.write(escribir(it, IVA))
            if 'cbtes_asoc' in dic:
                for it in dic['cbtes_asoc']:
                    it['tipo_reg'] = 3
                    archivo.write(escribir(it, CMP_ASOC))
            if 'opcionales' in dic:
                for it in dic['opcionales']:
                    it['tipo_reg'] = 6
                    archivo.write(escribir(it, OPCIONAL))

    if '/dbf' in sys.argv:
        formatos = [
            ('Encabezado', ENCABEZADO, encabezados),
            ('Tributo', TRIBUTO, dic.get('tributos', [])),
            ('Iva', IVA, dic.get('iva', [])),
            ('Comprobante Asociado', CMP_ASOC, dic.get('cbtes_asoc', [])),
            ('Datos Opcionales', OPCIONAL, dic.get("opcionales", [])),
        ]
        guardar_dbf(formatos, agrega, conf_dbf)


def depurar_xml(client, ruta="."):
    fecha = time.strftime("%Y%m%d%H%M%S")
    f = open(os.path.join(ruta, "request-%s.xml" % fecha), "w")
    f.write(client.xml_request)
    f.close()
    f = open(os.path.join(ruta, "response-%s.xml" % fecha), "w")
    f.write(client.xml_response)
    f.close()


if __name__ == "__main__":
    if '/ayuda' in sys.argv:
        print(LICENCIA)
        print("Opciones: ")
        print(" /ayuda: este mensaje")
        print(" /dummy: consulta estado de servidores")
        print(
            " /prueba: genera y autoriza una factura de prueba (no usar \
            en producción!)")
        print(" /ult: consulta último número de comprobante")
        print(
            " /debug: modo depuración (detalla y confirma las operaciones)")
        print(
            " /formato: muestra el formato de los archivos de entrada/salida")
        print(
            " /get: recupera datos de un comprobante autorizado \
            previamente (verificación)")
        print(
            " /xml: almacena los requerimientos y respuestas XML (depuración)")
        print(" /dbf: lee y almacena la información en tablas DBF")
        print(
            "Ver rece.ini para parámetros de configuración (URL, \
            certificados, etc.)")
        sys.exit(0)

    if '/debug'in sys.argv:
        DEBUG = True
        print("VERSION", __version__, "HOMO", HOMO)

    # si se pasa el archivo de configuración por parámetro, confirmar que
    # exista
    # y descartar que sea una opción
    if len(sys.argv) > 1 and (sys.argv[1][0] not in "-/"
                              or os.path.exists(sys.argv[1])):
        CONFIG_FILE = sys.argv.pop(1)
    if DEBUG:
        print("CONFIG_FILE:", CONFIG_FILE)

    config = SafeConfigParser()
    config.read(CONFIG_FILE)
    # print(CONFIG_FILE)
    cert = config.get('WSAA', 'CERT')
    privatekey = config.get('W SAA', 'PRIVATEKEY')
    cuit = config.get('WSFEv1', 'CUIT')
    if '/entrada' in sys.argv:
        entrada = sys.argv[sys.argv.index("/entrada")+1]
    else:
        entrada = config.get('WSFEv1', 'ENTRADA')
    salida = config.get('WSFEv1', 'SALIDA')

    if config.has_option('WSAA', 'URL') and not HOMO:
        wsaa_url = config.get('WSAA', 'URL')
    else:
        wsaa_url = None
    if config.has_option('WSFEv1', 'URL') and not HOMO:
        wsfev1_url = config.get('WSFEv1', 'URL')
    else:
        wsfev1_url = None

    if config.has_option('WSFEv1', 'REPROCESAR'):
        wsfev1_reprocesar = config.get('WSFEv1', 'REPROCESAR') == 'S'
    else:
        wsfev1_reprocesar = None

    if config.has_option('WSFEv1', 'XML_DIR'):
        wsfev1_xml_dir = config.get('WSFEv1', 'XML_DIR')
    else:
        wsfev1_xml_dir = "."

    if config.has_section('DBF'):
        conf_dbf = dict(config.items('DBF'))
        if DEBUG:
            print("conf_dbf", conf_dbf)
    else:
        conf_dbf = {}

    if config.has_section('PROXY') and not HOMO:
        proxy_dict = dict(
            ("proxy_%s" % k, v) for k, v in config.items('PROXY'))
        proxy_dict['proxy_port'] = int(proxy_dict['proxy_port'])
    else:
        proxy_dict = {}
    CACERT = config.has_option(
        'WSFEv1', 'CACERT') and config.get('WSFEv1', 'CACERT') or None
    WRAPPER = config.has_option(
        'WSFEv1', 'WRAPPER') and config.get('WSFEv1', 'WRAPPER') or None

    if '/xml'in sys.argv:
        XML = True

    RUTA_XML = config.has_option(
        'WSFEv1', 'XML') and config.get('WSFEv1', 'XML') or "."

    if DEBUG:
        print(
            "wsaa_url %s\nwsfev1_url %s\ncuit %s" % (
                wsaa_url, wsfev1_url, cuit)
        )
        if proxy_dict:
            print("proxy_dict=", proxy_dict)

    if '/x' in sys.argv:
        escribir_facturas(
            [{'err_msg': "Prueba", }], open("x.txt", "w"))

    try:
        ws = wsfev1.WSFEv1()
        ws.LanzarExcepciones = True
        ws.Conectar(
            "", wsfev1_url, proxy=proxy_dict, cacert=CACERT, wrapper=WRAPPER)
        ws.Cuit = cuit
        if wsfev1_reprocesar is not None:
            ws.Reprocesar = wsfev1_reprocesar

        if '/dummy' in sys.argv:
            print("Consultando estado de servidores...")
            ws.Dummy()
            print("AppServerStatus", ws.AppServerStatus)
            print("DbServerStatus", ws.DbServerStatus)
            print("AuthServerStatus", ws.AuthServerStatus)
            sys.exit(0)

        if '/formato' in sys.argv:
            print("Formato:")
            for msg, formato in [('Encabezado', ENCABEZADO),
                                 ('Tributo', TRIBUTO), ('Iva', IVA),
                                 ('Comprobante Asociado', CMP_ASOC),
                                 ('Opcionales', OPCIONAL)]:
                if '/dbf' not in sys.argv:
                    comienzo = 1
                    print("== %s ==" % msg)
                    for fmt in formato:
                        clave, longitud, tipo = fmt[0:3]
                        dec = len(fmt) > 3 and fmt[3] or (
                            tipo == 'I' and '2' or '')
                        print(" * Campo: %-20s Posición: %3d Longitud: %4d \
                            Tipo: %s Decimales: %s" % (
                                clave, comienzo, longitud, tipo, dec))
                        comienzo += longitud
                else:
                    from formatos.formato_dbf import definir_campos
                    filename = "%s.dbf" % msg.lower()[:8]
                    print("==== %s (%s) ====" % (msg, filename))
                    claves, campos = definir_campos(formato)
                    for campo in campos:
                        print(" * Campo: %s" % (campo,))
            sys.exit(0)

        # obteniendo el TA
        from wsaa import WSAA
        wsaa = WSAA()
        ta = wsaa.Autenticar(
            "wsfe", cert, privatekey, wsaa_url, proxy=proxy_dict,
            cacert=CACERT, wrapper=WRAPPER)
        if not ta:
            sys.exit("Imposible autenticar con WSAA: %s" % wsaa.Excepcion)
        ws.SetTicketAcceso(ta)

        if '/prueba' in sys.argv:
            # generar el archivo de prueba para la próxima factura
            tipo_cbte = 1
            punto_vta = 4002
            cbte_nro = ws.CompUltimoAutorizado(tipo_cbte, punto_vta)
            if not cbte_nro:
                cbte_nro = 0
            cbte_nro = int(cbte_nro)
            fecha = datetime.datetime.now().strftime("%Y%m%d")
            concepto = 1
            tipo_doc = 80
            nro_doc = "30628789661"
            cbt_desde = cbte_nro + 1
            cbt_hasta = cbte_nro + 1
            imp_total = "122.00"
            imp_tot_conc = "0.00"
            imp_neto = "100.00"
            imp_iva = "21.00"
            imp_trib = "1.00"
            imp_op_ex = "0.00"
            fecha_cbte = fecha
            fecha_venc_pago = None  # fecha
            # Fechas del período del servicio facturado (solo si concepto = 1?)
            fecha_serv_desde = ""
            fecha_serv_hasta = ""
            moneda_id = 'PES'
            moneda_ctz = '1.000'

            ws.CrearFactura(
                concepto, tipo_doc, nro_doc, tipo_cbte, punto_vta,
                cbt_desde, cbt_hasta, imp_total, imp_tot_conc, imp_neto,
                imp_iva, imp_trib, imp_op_ex, fecha_cbte, fecha_venc_pago,
                fecha_serv_desde, fecha_serv_hasta,
                moneda_id, moneda_ctz)

            if tipo_cbte not in (1, 2, 6, 7):
                tipo = 1
                pto_vta = 2
                nro = 1234
                ws.AgregarCmpAsoc(tipo, pto_vta, nro)

            if '--proyectos' in sys.argv:
                ws.AgregarOpcional(2, "1234")  # identificador del proyecto

            # datos opcionales para RG 3668 Impuesto al Valor Agregado -
            # Art.12:
            if '--rg3668' in sys.argv:
                ws.AgregarOpcional(5, "02")             # IVA Excepciones
                ws.AgregarOpcional(61, "80")            # Firmante Doc Tipo
                ws.AgregarOpcional(62, "20267565393")   # Firmante Doc Nro
                ws.AgregarOpcional(7, "01")             # Carácter del Firmante

            # RG 3.368 Establecimientos de educación pública de gestión privada
            if '--rg3749' in sys.argv:
                ws.AgregarOpcional(10, "1")             # Actividad Comprendida
                ws.AgregarOpcional(1011, "80")            # Tipo de Documento
                ws.AgregarOpcional(1012, "20267565393")   # Número de Documento

            tributo_id = 99
            desc = 'Impuesto Municipal Matanza'
            base_imp = 100
            alic = 1
            importe = 1
            ws.AgregarTributo(tributo_id, desc, base_imp, alic, importe)

            iva_id = 5  # 21%
            base_imp = 100
            importe = 21
            ws.AgregarIva(iva_id, base_imp, importe)

            f_entrada = open(entrada, "w")

            if DEBUG:
                print(ws.factura)

            dic = ws.factura
            escribir_facturas([dic], f_entrada, agrega=True)
            f_entrada.close()

        if '/ult' in sys.argv:
            print("Consultar ultimo numero:")
            i = sys.argv.index("/ult")
            if i+2 < len(sys.argv):
                tipo_cbte = int(sys.argv[i+1])
                punto_vta = int(sys.argv[i+2])
            else:
                tipo_cbte = int(raw_input("Tipo de comprobante: "))
                punto_vta = int(raw_input("Punto de venta: "))
            ult_cbte = ws.CompUltimoAutorizado(tipo_cbte, punto_vta)
            print("Ultimo numero: ", ult_cbte)
            print(ws.ErrMsg)
            depurar_xml(ws.client, RUTA_XML)
            escribir_facturas(
                [{'tipo_cbte': tipo_cbte,
                  'punto_vta': punto_vta,
                  'cbt_desde': ult_cbte,
                  'fecha_cbte': ws.FechaCbte,
                  'err_msg': ws.ErrMsg}],
                open(salida, "w")
            )
            sys.exit(0)

        if '/get' in sys.argv:
            print("Recuperar comprobante:")
            i = sys.argv.index("/get")
            if i+3 < len(sys.argv):
                tipo_cbte = int(sys.argv[i+1])
                punto_vta = int(sys.argv[i+2])
                cbte_nro = int(sys.argv[i+3])
            else:
                tipo_cbte = int(raw_input("Tipo de comprobante: "))
                punto_vta = int(raw_input("Punto de venta: "))
                cbte_nro = int(raw_input("Numero de comprobante: "))
            ws.CompConsultar(tipo_cbte, punto_vta, cbte_nro)

            ws.AnalizarXml("XmlResponse")
            print("FechaCbte = ", ws.FechaCbte)
            print("CbteNro = ", ws.CbteNro)
            print("PuntoVenta = ", ws.PuntoVenta)
            print("TipoDoc = ", ws.ObtenerTagXml('DocTipo'))
            print("NroDoc = ", ws.ObtenerTagXml('DocNro'))
            print("ImpTotal =", ws.ImpTotal)
            print("CAE = ", ws.CAE)
            print("Vencimiento = ", ws.Vencimiento)
            print("EmisionTipo = ", ws.EmisionTipo)
            print(ws.ErrMsg)

            depurar_xml(ws.client, RUTA_XML)
            # grabar todos los datos devueltos por AFIP:
            factura = ws.factura.copy()
            # actulizar los campos básicos:
            factura.update(
                {'tipo_cbte': tipo_cbte,
                 'punto_vta': ws.PuntoVenta,
                 'cbt_desde': ws.CbtDesde,
                 'cbt_hasta': ws.CbtHasta,
                 'fecha_cbte': ws.FechaCbte,
                 'tipo_doc': ws.ObtenerCampoFactura('tipo_doc'),
                 'nro_doc': ws.ObtenerCampoFactura('nro_doc'),
                 'imp_total': ws.ImpTotal,
                 'imp_neto': ws.ImpNeto,
                 'imp_iva': ws.ImpOpEx,
                 'imp_trib': ws.ImpTrib,
                 'imp_op_ex': ws.ImpTrib,
                 'cae': str(ws.CAE),
                 'fch_venc_cae': ws.Vencimiento,
                 'emision_tipo': ws.EmisionTipo,
                 'resultado': ws.Resultado,
                 'err_msg': ws.ErrMsg,
                 'motivos_obs': ws.Obs})
            escribir_facturas([factura], open(salida, "w"))

            sys.exit(0)

        if '/solicitarcaea' in sys.argv:
            i = sys.argv.index("/solicitarcaea")
            if i+2 < len(sys.argv):
                periodo = sys.argv[sys.argv.index("/solicitarcaea")+1]
                orden = sys.argv[sys.argv.index("/solicitarcaea")+2]
            else:
                periodo = raw_input("Periodo: ")
                orden = raw_input("Orden: ")

            if DEBUG:
                print(
                    "Solicitando CAEA para periodo %s orden %s" % (
                        periodo, orden)
                )

            caea = ws.CAEASolicitar(periodo, orden)
            print("CAEA:", caea)

            if ws.Errores:
                print("Errores:")
                for error in ws.Errores:
                    print(error)

            depurar_xml(ws.client, RUTA_XML)

            if not caea:
                if DEBUG:
                    print(
                        "Consultando CAEA para periodo %s orden %s" % (
                            periodo, orden)
                    )
                caea = ws.CAEAConsultar(periodo, orden)
                print("CAEA:", caea)

            if DEBUG:
                print("Periodo:", ws.Periodo)
                print("Orden:", ws.Orden)
                print("FchVigDesde:", ws.FchVigDesde)
                print("FchVigHasta:", ws.FchVigHasta)
                print("FchTopeInf:", ws.FchTopeInf)
                print("FchProceso:", ws.FchProceso)

            escribir_facturas(
                [{'cae': str(caea),
                  'emision_tipo': "CAEA"}],
                open(salida, "w")
            )

            sys.exit(0)

        if '/consultarcaea' in sys.argv:
            i = sys.argv.index("/consultarcaea")
            if i+2 < len(sys.argv):
                periodo = sys.argv[sys.argv.index("/consultarcaea")+1]
                orden = sys.argv[sys.argv.index("/consultarcaea")+2]
            else:
                periodo = raw_input("Periodo: ")
                orden = raw_input("Orden: ")

            if DEBUG:
                print(
                    "Consultando CAEA para periodo %s orden %s" % (
                        periodo, orden)
                )

            caea = ws.CAEAConsultar(periodo, orden)
            print("CAEA:", caea)

            if ws.Errores:
                print("Errores:")
                for error in ws.Errores:
                    print(error)

            if DEBUG:
                print("Periodo:", ws.Periodo)
                print("Orden:", ws.Orden)
                print("FchVigDesde:", ws.FchVigDesde)
                print("FchVigHasta:", ws.FchVigHasta)
                print("FchTopeInf:", ws.FchTopeInf)
                print("FchProceso:", ws.FchProceso)
            sys.exit(0)

        if '/ptosventa' in sys.argv:

            print("=== Puntos de Venta ===")
            print(u'\n'.join(ws.ParamGetPtosVenta()))
            sys.exit(0)

        if '/informarcaeanoutilizadoptovta' in sys.argv:
            i = sys.argv.index('/informarcaeanoutilizadoptovta')
            if i+2 < len(sys.argv):
                caea = sys.argv[i+1]
                pto_vta = sys.argv[i+2]
            else:
                caea = raw_input("CAEA: ")
                pto_vta = raw_input("Punto de Venta: ")
            if DEBUG:
                print(
                    "Informando CAEA no utilizado: %s pto_vta %s" % (
                        caea, pto_vta)
                )
            ok = ws.CAEASinMovimientoInformar(pto_vta, caea)
            print("Resultado:", ok)
            print("FchProceso:", ws.FchProceso)
            if ws.Errores:
                print("Errores:")
                for error in ws.Errores:
                    print(error)
            sys.exit(0)

        ws.LanzarExcepciones = False
        f_entrada = f_salida = None
        try:
            f_entrada = open(entrada, "r")
            f_salida = open(salida, "w")
            try:
                if DEBUG:
                    print("Autorizando usando entrada:", entrada)
                autorizar(ws, f_entrada, f_salida, '/informarcaea' in sys.argv)
            except SoapFault:
                XML = True
                raise
        finally:
            if f_entrada is not None:
                f_entrada.close()
            if f_salida is not None:
                f_salida.close()
            if XML:
                depurar_xml(ws.client, RUTA_XML)
        sys.exit(0)

    except SoapFault as e:
        print(
            "SoapFault:", e.faultcode, e.faultstring.encode("ascii", "ignore"))
        sys.exit(3)
    except Exception as e:
        e_str = unicode(e).encode("ascii", "ignore")
        if not e_str:
            e_str = repr(e)
        print("Excepcion:", e_str)
        escribir_facturas(
            [{'err_msg': e_str}],
            open(salida, "w")
        )
        ex = traceback.format_exception(
            sys.exc_type, sys.exc_value, sys.exc_traceback)
        open("traceback.txt", "wb").write('\n'.join(ex))

        if DEBUG:
            raise
        sys.exit(5)



#!/usr/bin/python
# -*- coding: utf-8 -*-
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTIBILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

"Módulo de Servidor COM para interfaz con Windows"

__author__ = "Mariano Reingart (reingart@gmail.com)"
__copyright__ = "Copyright (C) 2008 Mariano Reingart"
__license__ = "GPL 3.0"
__version__ = "1.31b"

import sys
import wsfe
import wsbfe
import wsfex
import wdigdepfiel
from php import (
    # SimpleXMLElement,
    SoapFault, SoapClient, parse_proxy
)
import traceback
# from httplib2 import socks
import warnings

try:
    from win32com.server.exception import COMException
    # import winerror
    # import pythoncom
except ImportError:
    print("La extension win32 no presente, interfaz COM deshabilitada")
    COMException = RuntimeError

HOMO = True

vbObjectError = -2147221504


def raiseSoapError(e):
    raise COMException(
        scode=vbObjectError,
        desc=unicode(e.faultstring),
        source=u"SOAP "+unicode(e.faultcode))


def checkError(result):
    if int(result.RError.percode) != 0:
        raise COMException(
            scode=vbObjectError + int(result.RError.percode),
            desc=unicode(result.RError.perrmsg), source="WebService")


def raisePythonException(e):
    ex = traceback.format_exception(
        sys.exc_type, sys.exc_value, sys.exc_traceback)
    raise COMException(
        scode=vbObjectError, desc=''.join(ex), source=u"Python")


class WSFE:
    "Interfaz para el WebService de Factura Electrónica"
    _public_methods_ = [
        'RecuperarQty', 'UltNro', 'RecuperaLastCMP', 'Aut', 'Dummy', 'Conectar'
    ]
    _public_attrs_ = ['Token', 'Sign', 'Cuit',
                      'AppServerStatus', 'DbServerStatus', 'AuthServerStatus',
                      'XmlRequest', 'XmlResponse', 'Version',
                      'Resultado', 'Motivo', 'Reproceso',
                      'CbtDesde', 'CbtHasta', 'FechaCbte', 'ImpTotal',
                      'ImpNeto', 'ImptoLiq', 'LastID', 'LastCMP', 'CAE',
                      'Vencimiento']
    _reg_progid_ = "WSFE"
    _reg_clsid_ = "{247A418E-BF48-48B4-B936-D6FF158C0168}"

    def __init__(self):
        self.Token = self.Sign = self.Cuit = None
        self.AppServerStatus = self.DbServerStatus = \
            self.AuthServerStatus = None
        self.XmlRequest = ''
        self.XmlResponse = ''
        self.Resultado = self.Motivo = self.Reproceso = ''
        self.LastID = self.LastCMP = self.CAE = self.Vencimiento = ''
        self.client = None
        self.Version = "%s %s" % (__version__, HOMO and 'Homologación' or '')
        self.CbtDesde = self.CbtHasta = self.FechaCbte = None
        self.ImpTotal = self.ImpNeto = self.ImptoLiq = None

    def Conectar(self, url="", proxy=""):
        "Establecer la conexión a los servidores de la AFIP"
        if HOMO or not url:
            url = wsfe.WSFEURL
        proxy_dict = parse_proxy(proxy)
        try:
            self.client = SoapClient(
                url,
                action=wsfe.SOAP_ACTION,
                namespace=wsfe.SOAP_NS,
                trace=False,
                exceptions=True,
                proxy=proxy_dict)
            return True
        except Exception as e:
            raisePythonException(e)

    def RecuperarQty(self):
        "Recuperar cantidad máxima de registros de detalle"
        try:
            results = self.client.FERecuperaQTYRequest(
                argAuth={
                    "Token": self.Token,
                    "Sign": self.Sign,
                    "cuit": long(self.Cuit)
                })
            checkError(results.FERecuperaQTYRequestResult)
            return int(results.FERecuperaQTYRequestResult.qty.value)
        except COMException:
            raise
        except Exception as e:
            print(e)
            ex = traceback.format_exception(
                sys.exc_type, sys.exc_value, sys.exc_traceback)
            raise RuntimeError(''.join(ex))

    def UltNro(self):
        "Recuperar el último número de transacción (id)"
        results = self.client.FEUltNroRequest(
            argAuth={
                'Token': self.Token,
                'Sign': self.Sign,
                'cuit': long(self.Cuit)
            })
        checkError(results.FEUltNroRequestResult)
        nro = results.FEUltNroRequestResult.nro.value
        self.LastID = str(nro)
        return str(nro)

    def RecuperaLastCMP(self, ptovta, tipocbte):
        "Recuperar último número de comprobante autorizado"
        results = self.client.FERecuperaLastCMPRequest(
            argAuth={
                'Token': self.Token,
                'Sign': self.Sign,
                'cuit': long(self.Cuit)
            },
            argTCMP={'PtoVta': ptovta, 'TipoCbte': tipocbte})
        checkError(results.FERecuperaLastCMPRequestResult)
        nro = results.FERecuperaLastCMPRequestResult.cbte_nro
        self.LastCMP = str(nro)
        return int(nro)

    def Aut(self, id, presta_serv, tipo_doc, nro_doc, tipo_cbte, punto_vta,
            cbt_desde, cbt_hasta, imp_total, imp_tot_conc, imp_neto,
            impto_liq, impto_liq_rni, imp_op_ex, fecha_cbte, fecha_venc_pago,
            fecha_serv_desde="", fecha_serv_hasta=""):
        "Autorizar la factura y devuelve CAE"
        try:
            detalle = {
                'tipo_doc': tipo_doc, 'nro_doc':  nro_doc,
                'tipo_cbte': tipo_cbte, 'punto_vta': punto_vta,
                'cbt_desde': cbt_desde, 'cbt_hasta': cbt_hasta,
                'imp_total': imp_total, 'imp_tot_conc': imp_tot_conc,
                'imp_neto': imp_neto, 'impto_liq': impto_liq,
                'impto_liq_rni': impto_liq_rni, 'imp_op_ex': imp_op_ex,
                'fecha_cbte': fecha_cbte,
                'fecha_venc_pago': fecha_venc_pago,
            }
            if fecha_serv_desde:
                detalle['fecha_serv_desde'] = fecha_serv_desde
            if fecha_serv_hasta:
                detalle['fecha_serv_hasta'] = fecha_serv_hasta

            results = self.client.FEAutRequest(
                argAuth={
                    'Token': self.Token,
                    'Sign': self.Sign,
                    'cuit': long(self.Cuit)
                }, Fer={'Fecr': {
                    'id': long(id),
                    'cantidadreg': 1,
                    'presta_serv': int(presta_serv)},
                    'Fedr': {'FEDetalleRequest': detalle, }})

            if 'FEAutRequestResult' not in results:
                return ''

            checkError(results.FEAutRequestResult)
            self.Resultado = str(results.FEAutRequestResult.resultado)
            # Motivo general:
            self.Motivo = str(results.FEAutRequestResult.motivo)
            # Motivo del detalle:
            motivo = str(
                results.FEAutRequestResult.FedResp.FEDetalleResponse.motivo)
            if self.Motivo in ("NULL", '00') and motivo not in ("NULL", "00"):
                self.Motivo = motivo
            self.Reproceso = str(results.FEAutRequestResult.reproceso)
            self.CAE = str(
                results.FEAutRequestResult.FedResp.FEDetalleResponse.cae)
            vto = str(
                results.FEAutRequestResult.FedResp.FEDetalleResponse.fecha_vto)
            self.Vencimiento = "%s/%s/%s" % (vto[6:8], vto[4:6], vto[0:4])

            # completo los campos devueltos
            self.CbtDesde = str(
                results.FEAutRequestResult.FedResp.FEDetalleResponse.cbt_desde)
            self.CbtHasta = str(
                results.FEAutRequestResult.FedResp.FEDetalleResponse.cbt_hasta)
            self.FechaCbte = str(
                results.FEAutRequestResult.FedResp.FEDetalleResponse.fecha_cbte
            )
            self.ImpTotal = str(
                results.FEAutRequestResult.FedResp.FEDetalleResponse.imp_total)
            self.ImpNeto = str(
                results.FEAutRequestResult.FedResp.FEDetalleResponse.imp_neto)
            self.ImptoLiq = str(
                results.FEAutRequestResult.FedResp.FEDetalleResponse.impto_liq)

            return self.CAE
        except SoapFault as e:
            raiseSoapError(e)
        except COMException:
            raise
        except Exception as e:
            raisePythonException(e)
        finally:
            # guardo los mensajes para depuración
            self.XmlRequest = self.client.xml_request
            self.XmlResponse = self.client.xml_response

    def Dummy(self):
        "Obtener el estado de los servidores de la AFIP"
        results = self.client.FEDummy()
        self.AppServerStatus = str(results.appserver)
        self.DbServerStatus = str(results.dbserver)
        self.AuthServerStatus = str(results.authserver)


class WSBFE:
    "Interfaz para el WebService de Bono Fiscal Electrónico (FE Bs. Capital)"
    _public_methods_ = ['CrearFactura', 'AgregarItem', 'Authorize', 'GetCMP',
                        'GetParamMon', 'GetParamTipoCbte', 'GetParamUMed',
                        'GetParamTipoIVA', 'GetParamNCM',
                        'Dummy', 'Conectar', 'GetLastCMP', 'GetLastID']
    _public_attrs_ = ['Token', 'Sign', 'Cuit',
                      'AppServerStatus', 'DbServerStatus', 'AuthServerStatus',
                      'XmlRequest', 'XmlResponse', 'Version',
                      'Resultado', 'Obs', 'Reproceso',
                      'Excepcion', 'LanzarExcepciones', 'Traceback',
                      'ErrCode', 'ErrMsg', 'CAE', 'Vencimiento', 'Eventos',
                      'FechaCbte', 'ImpNeto', 'ImptoLiq', 'ImpTotal']

    _reg_progid_ = "WSBFE"
    _reg_clsid_ = "{02CBC6DA-455D-4EE6-8302-411D13253CBF}"

    def __init__(self):
        self.Token = self.Sign = self.Cuit = None
        self.AppServerStatus = self.DbServerStatus = \
            self.AuthServerStatus = None
        self.XmlRequest = ''
        self.XmlResponse = ''
        self.Resultado = self.Motivo = self.Reproceso = ''
        self.LastID = self.LastCMP = self.CAE = self.Vencimiento = ''
        self.Obs = None
        self.client = None
        self.Version = "%s %s" % (__version__, HOMO and 'Homologación' or '')
        self.factura = None
        self.FechaCbte = ImpNeto = ImptoLiq = ImpTotal = None
        self.LanzarExcepciones = True
        self.Traceback = self.Excepcion = ""
        self.ErrCode = self.ErrMsg = ""

    def Conectar(self, url="", proxy=""):
        "Establecer la conexión a los servidores de la AFIP"
        if HOMO or not url:
            url = wsbfe.WSBFEURL
        proxy_dict = parse_proxy(proxy)
        try:
            # limpio errores
            self.Exception = self.Traceback = ""
            self.ErrCode = self.ErrMsg = ""
            self.client = SoapClient(
                url,
                action=wsbfe.SOAP_ACTION,
                namespace=wsbfe.SOAP_NS,
                trace=False,
                exceptions=True, proxy=proxy_dict)
            return True
        except Exception as e:
            if self.LanzarExcepciones:
                raisePythonException(e)
            else:
                ex = traceback.format_exception(
                    sys.exc_type, sys.exc_value, sys.exc_traceback)
                self.Traceback = ''.join(ex)
                self.Excepcion = traceback.format_exception_only(
                    sys.exc_type, sys.exc_value)[0]

    def CrearFactura(self, tipo_doc=80, nro_doc=23111111113,
                     zona=0, tipo_cbte=1, punto_vta=1, cbte_nro=0,
                     fecha_cbte=None, imp_total=0.0, imp_neto=0.0,
                     impto_liq=0.0, imp_tot_conc=0.0, impto_liq_rni=0.00,
                     imp_op_ex=0.00, imp_perc=0.00, imp_iibb=0.00,
                     imp_perc_mun=0.00, imp_internos=0.00, imp_moneda_id=0,
                     imp_moneda_ctz=1.0):
        "Creo un objeto factura (interna)"
        # Creo una factura para bonos fiscales electrónicos
        factura = wsbfe.FacturaBF()
        # Establezco el encabezado
        factura.tipo_doc = tipo_doc
        factura.nro_doc = nro_doc
        factura.zona = zona
        factura.tipo_cbte = tipo_cbte
        factura.punto_vta = punto_vta
        factura.cbte_nro = cbte_nro
        factura.fecha_cbte = fecha_cbte
        factura.imp_total = imp_total
        factura.imp_neto = imp_neto
        factura.impto_liq = impto_liq
        factura.imp_tot_conc = imp_tot_conc
        factura.impto_liq_rni = impto_liq_rni
        factura.imp_op_ex = imp_op_ex
        factura.imp_perc = imp_perc
        factura.imp_iibb = imp_iibb
        factura.imp_perc_mun = imp_perc_mun
        factura.imp_internos = imp_internos
        factura.imp_moneda_id = imp_moneda_id
        factura.imp_moneda_ctz = imp_moneda_ctz
        self.factura = factura

    def AgregarItem(
        self, ncm, sec, ds, qty, umed, precio, bonif, iva_id, imp_total
    ):
        "Agrego un item a una factura (interna)"
        # ds = unicode(ds, "latin1") # convierto a latin1
        # Nota: no se calcula neto, iva, etc (deben venir calculados!)
        item = wsbfe.ItemBF(
            ncm, sec, ds, qty, umed, precio, bonif, iva_id, imp_total)
        self.factura.items.append(item)

    def Authorize(self, id):
        "Autoriza la factura en memoria (interna)"
        try:
            # limpio errores
            self.Exception = self.Traceback = ""
            self.ErrCode = self.ErrMsg = ""
            # llamo al web service
            auth, events = wsbfe.authorize(
                self.client, self.Token, self.Sign, self.Cuit, id=id,
                factura=self.factura.to_dict())

            # Resultado: A: Aceptado, R: Rechazado
            self.Resultado = auth['resultado']
            # Obs:
            self.Obs = auth['obs'].strip(" ")
            self.Reproceso = auth['reproceso']
            self.CAE = auth['cae']
            vto = str(auth['fch_venc_cae'])
            self.Vencimiento = "%s/%s/%s" % (vto[6:8], vto[4:6], vto[0:4])
            self.Eventos = [
                '%s: %s' % (evt['code'], evt['msg']) for evt in events]
            return self.CAE
        except wsbfe.BFEError as e:
            self.ErrCode = unicode(e.code)
            self.ErrMsg = unicode(e.msg)
            if self.LanzarExcepciones:
                raise COMException(scode=vbObjectError + int(e.code),
                                   desc=unicode(e.msg), source="WebService")
        except SoapFault as e:
            self.ErrCode = unicode(e.faultcode)
            self.ErrMsg = unicode(e.faultstring)
            if self.LanzarExcepciones:
                raiseSoapError(e)
        except COMException:
            if self.LanzarExcepciones:
                raise
        except Exception as e:
            if self.LanzarExcepciones:
                raisePythonException(e)
            else:
                ex = traceback.format_exception(
                    sys.exc_type, sys.exc_value, sys.exc_traceback)
                self.Traceback = ''.join(ex)
                self.Excepcion = traceback.format_exception_only(
                    sys.exc_type, sys.exc_value)[0]
        finally:
            # guardo los mensajes para depuración
            self.XmlRequest = self.client.xml_request
            self.XmlResponse = self.client.xml_response

    def Dummy(self):
        "Obtener el estado de los servidores de la AFIP"
        results = wsbfe.dummy(self.client)
        self.AppServerStatus = str(results['appserver'])
        self.DbServerStatus = str(results['dbserver'])
        self.AuthServerStatus = str(results['authserver'])

    def GetCMP(self, tipo_cbte, punto_vta, cbte_nro):
        try:
            # limpio errores
            self.Exception = self.Traceback = ""
            self.ErrCode = self.ErrMsg = ""
            cbt, events = wsbfe.get_cmp(self.client,
                                        self.Token, self.Sign, self.Cuit,
                                        tipo_cbte, punto_vta, cbte_nro)

            self.FechaCbte = cbt['fch_cbte']
            self.ImpTotal = cbt['imp_total']
            self.ImpNeto = cbt['imp_neto']
            self.ImptoLiq = cbt['impto_liq']

            # Obs, cae y fecha cae
            self.Obs = cbt['obs'].strip(" ")
            self.CAE = cbt['cae']
            vto = str(cbt['fch_venc_cae'])
            self.Vencimiento = "%s/%s/%s" % (vto[6:8], vto[4:6], vto[0:4])
            self.Eventos = [
                '%s: %s' % (evt['code'], evt['msg']) for evt in events]
            return self.CAE

        except wsbfe.BFEError as e:
            self.ErrCode = unicode(e.code)
            self.ErrMsg = unicode(e.msg)
            if self.LanzarExcepciones:
                raise COMException(scode=vbObjectError + int(e.code),
                                   desc=unicode(e.msg), source="WebService")
        except SoapFault as e:
            self.ErrCode = unicode(e.faultcode)
            self.ErrMsg = unicode(e.faultstring)
            if self.LanzarExcepciones:
                raiseSoapError(e)
        except COMException:
            if self.LanzarExcepciones:
                raise
        except Exception as e:
            if self.LanzarExcepciones:
                raisePythonException(e)
            else:
                ex = traceback.format_exception(
                    sys.exc_type, sys.exc_value, sys.exc_traceback)
                self.Traceback = ''.join(ex)
                self.Excepcion = traceback.format_exception_only(
                    sys.exc_type, sys.exc_value)[0]
        finally:
            # guardo los mensajes para depuración
            self.XmlRequest = self.client.xml_request
            self.XmlResponse = self.client.xml_response

    def Factura(self):
        return self.factura

    def GetLastCMP(self, tipo_cbte, punto_vta):
        "Recuperar último número de comprobante emitido"
        try:
            # limpio errores
            self.Exception = self.Traceback = ""
            self.ErrCode = self.ErrMsg = ""
            cbte_nro, cbte_fecha, events = wsbfe.get_last_cmp(
                self.client, self.Token, self.Sign, self.Cuit,
                tipo_cbte, punto_vta)
            return cbte_nro
        except wsbfe.BFEError as e:
            self.ErrCode = unicode(e.code)
            self.ErrMsg = unicode(e.msg)
            if self.LanzarExcepciones:
                raise COMException(scode=vbObjectError + int(e.code),
                                   desc=unicode(e.msg), source="WebService")
        except SoapFault as e:
            self.ErrCode = unicode(e.faultcode)
            self.ErrMsg = unicode(e.faultstring)
            if self.LanzarExcepciones:
                raiseSoapError(e)
        except COMException:
            if self.LanzarExcepciones:
                raise
        except Exception as e:
            if self.LanzarExcepciones:
                raisePythonException(e)
            else:
                ex = traceback.format_exception(
                    sys.exc_type, sys.exc_value, sys.exc_traceback)
                self.Traceback = ''.join(ex)
                self.Excepcion = traceback.format_exception_only(
                    sys.exc_type, sys.exc_value)[0]
        finally:
            # guardo los mensajes para depuración
            self.XmlRequest = self.client.xml_request
            self.XmlResponse = self.client.xml_response

    def GetLastID(self):
        "Recuperar último número de transacción (ID)"
        try:
            # limpio errores
            self.Exception = self.Traceback = ""
            self.ErrCode = self.ErrMsg = ""
            id, events = wsbfe.get_last_id(self.client,
                                           self.Token, self.Sign, self.Cuit)
            return id
        except wsbfe.BFEError as e:
            self.ErrCode = unicode(e.code)
            self.ErrMsg = unicode(e.msg)
            if self.LanzarExcepciones:
                raise COMException(scode=vbObjectError + int(e.code),
                                   desc=unicode(e.msg), source="WebService")
        except SoapFault as e:
            self.ErrCode = unicode(e.faultcode)
            self.ErrMsg = unicode(e.faultstring)
            if self.LanzarExcepciones:
                raiseSoapError(e)
        except COMException:
            if self.LanzarExcepciones:
                raise
        except Exception as e:
            if self.LanzarExcepciones:
                raisePythonException(e)
            else:
                ex = traceback.format_exception(
                    sys.exc_type, sys.exc_value, sys.exc_traceback)
                self.Traceback = ''.join(ex)
                self.Excepcion = traceback.format_exception_only(
                    sys.exc_type, sys.exc_value)[0]
        finally:
            # guardo los mensajes para depuración
            self.XmlRequest = self.client.xml_request
            self.XmlResponse = self.client.xml_response

    def GetParamMon(self):
        "Recuperar lista de valores referenciales de codigos de Moneda"
        params = wsbfe.get_param_mon(
            self.client, self.Token, self.Sign, self.Cuit)
        return [
            '%(id)s: %(ds)s (%(vig_desde)s - %(vig_hasta)s)' % p for p in
            params]

    def GetParamTipoCbte(self):
        "Recuperar lista de valores referenciales de códigos de Tipos de \
        comprobante"
        params = wsbfe.get_param_tipo_cbte(
            self.client, self.Token, self.Sign, self.Cuit)
        return [
            '%(id)s: %(ds)s (%(vig_desde)s - %(vig_hasta)s)' % p for p in
            params]

    def GetParamUMed(self):
        "Recuperar lista de valores referenciales de Unidades de Medidas"
        params = wsbfe.get_param_umed(
            self.client, self.Token, self.Sign, self.Cuit)
        return [
            '%(id)s: %(ds)s (%(vig_desde)s - %(vig_hasta)s)' % p for p in
            params]

    def GetParamTipoIVA(self):
        "Recuperar lista de valores referenciales de tipos de IVA (alícuotas)"
        params = wsbfe.get_param_tipo_iva(
            self.client, self.Token, self.Sign, self.Cuit)
        return [
            '%(id)s: %(ds)s (%(vig_desde)s - %(vig_hasta)s)' % p for p in
            params]

    def GetParamNCM(self):
        "Recuperar lista de valores referenciales de códigos del Nomenclador \
        Común del Mercosur"
        params = wsbfe.get_param_ncm(
            self.client, self.Token, self.Sign, self.Cuit)
        return [
            '%(codigo)s: %(ds)s (%(vig_desde)s - %(vig_hasta)s)' % p for p in
            params]


class WSFEX:
    "Interfaz para el WebService de Factura Electrónica Exportación"
    _public_methods_ = ['CrearFactura', 'AgregarItem', 'Authorize', 'GetCMP',
                        'AgregarPermiso', 'AgregarCmpAsoc',
                        'GetParamMon', 'GetParamTipoCbte', 'GetParamTipoExpo',
                        'GetParamIdiomas', 'GetParamUMed', 'GetParamIncoterms',
                        'GetParamDstPais', 'GetParamDstCUIT',
                        'GetParamCtz', 'Dummy', 'Conectar', 'GetLastCMP',
                        'GetLastID']
    _public_attrs_ = ['Token', 'Sign', 'Cuit',
                      'AppServerStatus', 'DbServerStatus', 'AuthServerStatus',
                      'XmlRequest', 'XmlResponse', 'Version',
                      'Resultado', 'Obs', 'Reproceso',
                      'CAE', 'Vencimiento', 'Eventos', 'ErrCode', 'ErrMsg',
                      'Excepcion', 'LanzarExcepciones', 'Traceback',
                      'CbteNro', 'FechaCbte', 'ImpTotal']

    _reg_progid_ = "WSFEX"
    _reg_clsid_ = "{B3C8D3D3-D5DA-44C9-B003-11845803B2BD}"

    def __init__(self):
        self.Token = self.Sign = self.Cuit = None
        self.AppServerStatus = self.DbServerStatus = \
            self.AuthServerStatus = None
        self.XmlRequest = ''
        self.XmlResponse = ''
        self.Resultado = self.Motivo = self.Reproceso = ''
        self.LastID = self.LastCMP = self.CAE = self.Vencimiento = ''
        self.client = None
        self.Version = "%s %s" % (__version__, HOMO and 'Homologación' or '')
        self.factura = None
        self.CbteNro = self.FechaCbte = ImpTotal = None
        self.ErrCode = self.ErrMsg = None
        self.LanzarExcepciones = True
        self.Traceback = self.Excepcion = ""

    def Conectar(self, url="", proxy=""):
        "Establecer la conexión a los servidores de la AFIP"
        if HOMO or not url:
            url = wsfex.WSFEXURL
        proxy_dict = parse_proxy(proxy)
        try:
            # limpio errores
            self.Exception = self.Traceback = ""
            self.ErrCode = self.ErrMsg = ""
            self.client = SoapClient(url,
                                     action=wsfex.SOAP_ACTION,
                                     namespace=wsfex.SOAP_NS,
                                     trace=False,
                                     exceptions=True, proxy=proxy_dict)
            return True
        except Exception as e:
            raisePythonException(e)

    def CrearFactura(self, tipo_cbte=19, punto_vta=1, cbte_nro=0,
                     fecha_cbte=None, imp_total=0.0, tipo_expo=1,
                     permiso_existente="N", dst_cmp=None,
                     cliente="", cuit_pais_cliente="", domicilio_cliente="",
                     id_impositivo="", moneda_id="PES", moneda_ctz=1.0,
                     obs_comerciales="", obs="", forma_pago="", incoterms="",
                     idioma_cbte=7):
        "Creo un objeto factura (interna)"
        # Creo una factura electronica de exportación
        factura = wsfex.FacturaEX()
        # Establezco el encabezado
        factura.tipo_cbte = tipo_cbte
        factura.punto_vta = punto_vta
        factura.cbte_nro = cbte_nro
        factura.fecha_cbte = fecha_cbte
        factura.tipo_expo = tipo_expo
        factura.permiso_existente = permiso_existente
        factura.dst_cmp = dst_cmp
        factura.cliente = cliente
        factura.cuit_pais_cliente = cuit_pais_cliente
        factura.domicilio_cliente = domicilio_cliente
        factura.id_impositivo = id_impositivo
        factura.moneda_id = moneda_id
        factura.moneda_ctz = moneda_ctz
        factura.obs_comerciales = obs_comerciales
        factura.obs = obs
        factura.forma_pago = forma_pago
        factura.incoterms = incoterms
        factura.idioma_cbte = idioma_cbte
        factura.imp_total = imp_total
        self.factura = factura

    def AgregarItem(self, codigo, ds, qty, umed, precio, imp_total):
        "Agrego un item a una factura (interna)"
        # Nota: no se calcula total (debe venir calculado!)
        item = wsfex.ItemFEX(codigo, ds, qty, umed, precio, imp_total)
        self.factura.items.append(item)

    def AgregarPermiso(self, id, dst):
        "Agrego un permiso a una factura (interna)"
        permiso = wsfex.PermisoFEX(id, dst)
        self.factura.add_permiso(permiso)

    def AgregarCmpAsoc(self, tipo=19, punto_vta=0, cbte_nro=0):
        "Agrego un comprobante asociado a una factura (interna)"
        cmp_asoc = wsfex.CmpAsocFEX(tipo, punto_vta, cbte_nro)
        self.factura.add_cmp_asoc(cmp_asoc)

    def Authorize(self, id):
        "Autoriza la factura cargada en memoria"
        try:
            # limpio errores
            self.Exception = self.Traceback = ""
            self.ErrCode = self.ErrMsg = ""
            # llamo al web service
            auth, events = wsfex.authorize(self.client,
                                           self.Token, self.Sign, self.Cuit,
                                           id=id,
                                           factura=self.factura.to_dict())

            # Resultado: A: Aceptado, R: Rechazado
            self.Resultado = auth['resultado']
            # Obs:
            self.Obs = auth['obs'].strip(" ")
            self.Reproceso = auth['reproceso']
            self.CAE = auth['cae']
            self.CbteNro = auth['cbte_nro']
            vto = str(auth['fch_venc_cae'])
            self.Vencimiento = "%s/%s/%s" % (vto[6:8], vto[4:6], vto[0:4])
            self.Eventos = [
                '%s: %s' % (evt['code'], evt['msg']) for evt in events]
            return self.CAE
        except wsfex.FEXError as e:
            self.ErrCode = unicode(e.code)
            self.ErrMsg = unicode(e.msg)
            if self.LanzarExcepciones:
                raise COMException(scode=vbObjectError + int(e.code),
                                   desc=unicode(e.msg), source="WebService")
        except SoapFault as e:
            self.ErrCode = unicode(e.faultcode)
            self.ErrMsg = unicode(e.faultstring)
            if self.LanzarExcepciones:
                raiseSoapError(e)
        except COMException:
            if self.LanzarExcepciones:
                raise
        except Exception as e:
            if self.LanzarExcepciones:
                raisePythonException(e)
            else:
                ex = traceback.format_exception(
                    sys.exc_type, sys.exc_value, sys.exc_traceback)
                self.Traceback = ''.join(ex)
                self.Excepcion = traceback.format_exception_only(
                    sys.exc_type, sys.exc_value)[0]
        finally:
            # guardo datos de depuración
            self.XmlRequest = self.client.xml_request
            self.XmlResponse = self.client.xml_response

    def Dummy(self):
        "Obtener el estado de los servidores de la AFIP"
        results = wsfex.dummy(self.client)
        self.AppServerStatus = str(results['appserver'])
        self.DbServerStatus = str(results['dbserver'])
        self.AuthServerStatus = str(results['authserver'])

    def GetCMP(self, tipo_cbte, punto_vta, cbte_nro):
        "Recuperar los datos completos de un comprobante ya autorizado"
        try:
            # limpio errores
            self.Exception = self.Traceback = ""
            self.ErrCode = self.ErrMsg = ""
            # llamo al web service
            cbt, events = wsfex.get_cmp(self.client,
                                        self.Token, self.Sign, self.Cuit,
                                        tipo_cbte, punto_vta, cbte_nro)

            self.FechaCbte = cbt['fch_cbte']
            self.ImpTotal = cbt['imp_total']

            # Obs, cae y fecha cae
            self.Obs = cbt['obs'].strip(" ")
            self.CAE = cbt['cae']
            vto = str(cbt['fch_venc_cae'])
            self.Vencimiento = "%s/%s/%s" % (vto[6:8], vto[4:6], vto[0:4])

            self.Eventos = [
                '%s: %s' % (evt['code'], evt['msg']) for evt in events]
            return self.CAE

        except wsfex.FEXError as e:
            self.ErrCode = unicode(e.code)
            self.ErrMsg = unicode(e.msg)
            if self.LanzarExcepciones:
                raise COMException(scode=vbObjectError + int(e.code),
                                   desc=unicode(e.msg), source="WebService")
        except SoapFault as e:
            self.ErrCode = unicode(e.faultcode)
            self.ErrMsg = unicode(e.faultstring)
            if self.LanzarExcepciones:
                raiseSoapError(e)
        except Exception as e:
            if self.LanzarExcepciones:
                raisePythonException(e)
            else:
                ex = traceback.format_exception(
                    sys.exc_type, sys.exc_value, sys.exc_traceback)
                self.Traceback = ''.join(ex)
                try:
                    self.Excepcion = u"%s" % (e)
                except Exception as e:
                    pass
        finally:
            # guardo datos de depuración
            self.XmlRequest = self.client.xml_request
            self.XmlResponse = self.client.xml_response

    def Factura(self):
        return self.factura

    def GetLastCMP(self, tipo_cbte, punto_vta):
        "Recuperar último número de comprobante emitido"
        try:
            # limpio errores
            self.Exception = self.Traceback = ""
            self.ErrCode = self.ErrMsg = ""
            # llamo al web service
            cbte_nro, cbte_fecha, events = wsfex.get_last_cmp(self.client,
                                                              self.Token,
                                                              self.Sign,
                                                              self.Cuit,
                                                              tipo_cbte,
                                                              punto_vta)
            return cbte_nro
        except wsfex.FEXError as e:
            self.ErrCode = unicode(e.code)
            self.ErrMsg = unicode(e.msg)
            if self.LanzarExcepciones:
                raise COMException(scode=vbObjectError + int(e.code),
                                   desc=unicode(e.msg), source="WebService")
        except SoapFault as e:
            self.ErrCode = unicode(e.faultcode)
            self.ErrMsg = unicode(e.faultstring)
            if self.LanzarExcepciones:
                raiseSoapError(e)
        except Exception as e:
            if self.LanzarExcepciones:
                raisePythonException(e)
            else:
                ex = traceback.format_exception(
                    sys.exc_type, sys.exc_value, sys.exc_traceback)
                self.Traceback = ''.join(ex)
                try:
                    self.Excepcion = u"%s" % (e)
                except Exception as e:
                    pass
        finally:
            # guardo datos de depuración
            self.XmlRequest = self.client.xml_request
            self.XmlResponse = self.client.xml_response

    def GetLastID(self):
        "Recuperar último número de transacción (ID)"
        try:
            # limpio errores
            self.Exception = self.Traceback = ""
            self.ErrCode = self.ErrMsg = ""
            # llamo al web service
            id, events = wsfex.get_last_id(self.client,
                                           self.Token, self.Sign, self.Cuit)
            return id
        except wsfex.FEXError as e:
            self.ErrCode = unicode(e.code)
            self.ErrMsg = unicode(e.msg)
            if self.LanzarExcepciones:
                raise COMException(scode=vbObjectError + int(e.code),
                                   desc=unicode(e.msg), source="WebService")
        except SoapFault as e:
            self.ErrCode = unicode(e.faultcode)
            self.ErrMsg = unicode(e.faultstring)
            if self.LanzarExcepciones:
                raiseSoapError(e)
        except Exception as e:
            if self.LanzarExcepciones:
                raisePythonException(e)
            else:
                ex = traceback.format_exception(
                    sys.exc_type, sys.exc_value, sys.exc_traceback)
                self.Traceback = ''.join(ex)
                try:
                    self.Excepcion = u"%s" % (e)
                except Exception as e:
                    pass
        finally:
            # guardo datos de depuración
            self.XmlRequest = self.client.xml_request
            self.XmlResponse = self.client.xml_response

    def GetParamMon(self):
        "Recuperar lista de valores referenciales de codigos de Moneda"
        params = wsfex.get_param_mon(
            self.client, self.Token, self.Sign, self.Cuit)
        return [
            '%(id)s: %(ds)s (%(vig_desde)s - %(vig_hasta)s)' % p for p in
            params]

    def GetParamTipoCbte(self):
        "Recuperar lista de valores referenciales de códigos de Tipos de \
        comprobante"
        params = wsfex.get_param_tipo_cbte(
            self.client, self.Token, self.Sign, self.Cuit)
        return [
            '%(id)s: %(ds)s (%(vig_desde)s - %(vig_hasta)s)' % p for p in
            params]

    def GetParamTipoExpo(self):
        "Recuperar lista de valores referenciales de códigos de Tipo de \
        exportación"
        params = wsfex.get_param_tipo_expo(
            self.client, self.Token, self.Sign, self.Cuit)
        return [
            '%(id)s: %(ds)s (%(vig_desde)s - %(vig_hasta)s)' % p for p in
            params]

    def GetParamIdiomas(self):
        "Recuperar lista de valores referenciales de códigos de Idiomas"
        params = wsfex.get_param_idiomas(
            self.client, self.Token, self.Sign, self.Cuit)
        return [
            '%(id)s: %(ds)s (%(vig_desde)s - %(vig_hasta)s)' % p for p in
            params]

    def GetParamUMed(self):
        "Recuperar lista de valores referenciales de Unidades de Medidas"
        params = wsfex.get_param_umed(
            self.client, self.Token, self.Sign, self.Cuit)
        return [
            '%(id)s: %(ds)s (%(vig_desde)s - %(vig_hasta)s)' % p for p in
            params]

    def GetParamIncoterms(self):
        "Recuperar lista de valores referenciales de Incoterms"
        params = wsfex.get_param_incoterms(
            self.client, self.Token, self.Sign, self.Cuit)
        return [
            '%(id)s: %(ds)s (%(vig_desde)s - %(vig_hasta)s)' % p for p in
            params]

    def GetParamDstPais(self):
        "Recuperar lista de valores referenciales de códigos de Países"
        params = wsfex.get_param_dst_pais(
            self.client, self.Token, self.Sign, self.Cuit)
        return ['%s: %s' % (p['codigo'], p['ds']) for p in params]

    def GetParamDstCUIT(self):
        "Recuperar lista de valores referenciales de CUIT de Países"
        params = wsfex.get_param_dst_cuit(
            self.client, self.Token, self.Sign, self.Cuit)
        return ['%s: %s' % (p['cuit'], p['ds']) for p in params]

    def GetParamCtz(self, moneda_id='DOL'):
        "Recuperar cotización de moneda"
        try:
            param = wsfex.get_param_ctz(
                self.client, self.Token, self.Sign, self.Cuit, moneda_id)
            return "%(fecha)s: %(ctz)s" % param
        except wsfex.FEXError as e:
            self.ErrCode = unicode(e.code)
            self.ErrMsg = unicode(e.msg)
            raise COMException(scode=vbObjectError + int(e.code),
                               desc=unicode(e.msg), source="WebService")
        except SoapFault as e:
            self.ErrCode = unicode(e.faultcode)
            self.ErrMsg = unicode(e.faultstring)
            raiseSoapError(e)
        except COMException:
            raise
        except Exception as e:
            raisePythonException(e)
        finally:
            # guardo datos de depuración
            self.XmlRequest = self.client.xml_request
            self.XmlResponse = self.client.xml_response


class WSCTG:
    "Interfaz para el WebService de Código de Trazabilidad de Granos"
    _public_methods_ = ['Conectar', 'Dummy', 'SolicitarCTG', 'ConfirmarCTG']
    _public_attrs_ = ['Token', 'Sign', 'Cuit',
                      'AppServerStatus', 'DbServerStatus', 'AuthServerStatus',
                      'XmlRequest', 'XmlResponse', 'Version', 'NumeroCTG',
                      'CodigoTransaccion', 'Observaciones']
    _reg_progid_ = "WSCTG"
    _reg_clsid_ = "{8F1CE1EA-6D7A-45E7-A5BB-B17E2EB1F965}"

    def __init__(self):
        self.Token = self.Sign = self.Cuit = None
        self.AppServerStatus = self.DbServerStatus = \
            self.AuthServerStatus = None
        self.XmlRequest = ''
        self.XmlResponse = ''
        self.CodError = self.DescError = ''
        self.client = None
        self.Version = "%s %s" % (__version__, HOMO and 'Homologación' or '')
        self.NumeroCTG = ''
        self.CodigoTransaccion = self.Observaciones = ''

    def Conectar(self, url="", proxy=""):
        "Establecer la conexión a los servidores de la AFIP"
        if HOMO or not url:
            url = wsctg.WSCTGURL
        proxy_dict = parse_proxy(proxy)
        try:
            self.client = SoapClient(url,
                                     action=wsctg.SOAP_ACTION,
                                     namespace=wsctg.SOAP_NS,
                                     trace=False, ns='ctg', soap_ns='soapenv',
                                     exceptions=True, proxy=proxy_dict)
            return True
        except Exception as e:
            raisePythonException(e)

    def Dummy(self):
        "Obtener el estado de los servidores de la AFIP"
        try:
            results = wsctg.dummy(self.client)
            self.AppServerStatus = str(results['appserver'])
            self.DbServerStatus = str(results['dbserver'])
            self.AuthServerStatus = str(results['authserver'])
        except SoapFault as e:
            raiseSoapError(e)
        except COMException:
            raise
        except Exception as e:
            raisePythonException(e)
        finally:
            # guardo datos de depuración
            self.XmlRequest = self.client.xml_request
            self.XmlResponse = self.client.xml_response

    def SolicitarCTG(self, numero_carta_de_porte, codigo_especie,
                     cuit_remitente_comercial, cuit_destino, cuit_destinatario,
                     codigo_localidad_origen, codigo_localidad_destino,
                     codigo_cosecha, peso_neto_carga, cant_horas,
                     patente_vehiculo, cuit_transportista):
        "Obtener Número de CTG"
        try:
            ret = wsctg.solicitar_ctg(
                self.client, self.Token, self.Sign,
                self.Cuit,
                numeroCartaDePorte=numero_carta_de_porte,
                codigoEspecie=codigo_especie,
                cuitRemitenteComercial=cuit_remitente_comercial,
                cuitDestino=cuit_destino,
                cuitDestinatario=cuit_destinatario,
                codigoLocalidadOrigen=codigo_localidad_origen,
                codigoLocalidadDestino=codigo_localidad_destino,
                codigoCosecha=codigo_cosecha,
                pesoNetoCarga=peso_neto_carga, cantHoras=cant_horas,
                patenteVehiculo=patente_vehiculo,
                cuitTransportista=cuit_transportista)
            self.NumeroCTG = ret
            return self.NumeroCTG

        except SoapFault as e:
            raiseSoapError(e)
        except COMException:
            raise
        except Exception as e:
            raisePythonException(e)
        finally:
            # guardo datos de depuración
            self.XmlRequest = self.client.xml_request
            self.XmlResponse = self.client.xml_response

    def ConfirmarCTG(self, numero_carta_de_porte, numero_CTG, cuit_transportista, peso_neto_carga):
        "Confirma número de CTG"
        try:
            ret = wsctg.confirmar_ctg(
                self.client, self.Token, self.Sign, self.Cuit,
                numeroCartaDePorte=numero_carta_de_porte,
                numeroCTG=numero_CTG,
                cuitTransportista=cuit_transportista,
                pesoNetoCarga=peso_neto_carga)
            self.CodigoTransaccion, self.Observaciones = ret
            return self.CodigoTransaccion
        except SoapFault as e:
            raiseSoapError(e)
        except COMException:
            raise
        except Exception as e:
            raisePythonException(e)
        finally:
            # guardo datos de depuración
            self.XmlRequest = self.client.xml_request
            self.XmlResponse = self.client.xml_response


class wDigDepFiel:
    "Interfaz para el WebService de Depositario Fiel"
    _public_methods_ = [
        'Conectar', 'Dummy', 'AvisoDigit', 'AvisoRecepAcept',
        'IniciarAviso', 'AgregarFamilia']
    _public_attrs_ = ['Token', 'Sign', 'Cuit',
                      'AppServerStatus', 'DbServerStatus', 'AuthServerStatus',
                      'XmlRequest', 'XmlResponse', 'Version',
                      'CodError', 'DescError']
    _reg_progid_ = "wDigDepFiel"
    _reg_clsid_ = "{20516CB7-2B85-4562-8821-8F9D699CCAB6}"

    def __init__(self):
        self.Token = self.Sign = self.Cuit = None
        self.AppServerStatus = self.DbServerStatus = \
            self.AuthServerStatus = None
        self.XmlRequest = ''
        self.XmlResponse = ''
        self.CodError = self.DescError = ''
        self.client = None
        self.Version = "%s %s" % (__version__, HOMO and 'Homologación' or '')

    def Conectar(self, url="", proxy=""):
        "Establecer la conexión a los servidores de la AFIP"
        if HOMO or not url:
            url = wdigdepfiel.WSDDFURL
        proxy_dict = parse_proxy(proxy)
        try:
            self.client = SoapClient(url,
                                     action=wdigdepfiel.SOAP_ACTION,
                                     namespace=wdigdepfiel.SOAP_NS,
                                     trace=False, ns='ar', soap_ns='soap',
                                     exceptions=True, proxy=proxy_dict)
            return True
        except Exception as e:
            raisePythonException(e)

    def Dummy(self):
        "Obtener el estado de los servidores de la AFIP"
        results = wdigdepfiel.dummy(self.client)
        self.AppServerStatus = str(results['appserver'])
        self.DbServerStatus = str(results['dbserver'])
        self.AuthServerStatus = str(results['authserver'])

    def AvisoRecepAcept(self, tipo_agente, rol,
                        nro_legajo, cuit_declarante, cuit_psad, cuit_ie,
                        codigo, fecha_hora_acept, ticket):
        "Aviso de Recepción y Aceptación"
        try:
            ret = wdigdepfiel.aviso_recep_acept(
                self.client, self.Token, self.Sign, self.Cuit, tipo_agente,
                rol, nro_legajo, cuit_declarante, cuit_psad, cuit_ie,
                codigo, fecha_hora_acept, ticket)

            self.CodError, self.DescError = ret
            return self.CodError

        except SoapFault as e:
            raiseSoapError(e)
        except COMException:
            raise
        except Exception as e:
            raisePythonException(e)
        finally:
            # guardo datos de depuración
            self.XmlRequest = self.client.xml_request
            self.XmlResponse = self.client.xml_response

    def IniciarAviso(self):
        "Inicializar familias para Aviso de Digitalización"
        self.familias = []

    def AgregarFamilia(self, codigo, cantidad):
        "Agregar famila para Aviso de Digitalización"
        self.familias.append(
            {'Familia': {'codigo': codigo, 'cantidad': cantidad}})

    def AvisoDigit(self, tipo_agente, rol,
                   nro_legajo, cuit_declarante, cuit_psad, cuit_ie, cuit_ata,
                   codigo, url, ticket, hashing, cantidad_total):
        "Aviso de Digitalización"
        try:
            familias = self.familias
            ret = wdigdepfiel.aviso_digit(
                self.client, self.Token, self.Sign, self.Cuit, tipo_agente,
                rol, nro_legajo, cuit_declarante, cuit_psad, cuit_ie, cuit_ata,
                codigo, url, familias, ticket, hashing, cantidad_total)

            self.CodError, self.DescError = ret
            return self.CodError

        except SoapFault as e:
            raiseSoapError(e)
        except COMException:
            raise
        except Exception as e:
            raisePythonException(e)
        finally:
            # guardo datos de depuración
            self.XmlRequest = self.client.xml_request
            self.XmlResponse = self.client.xml_response


if __name__ == '__main__':

    warnings.warn(
        "Este modulo es obsoleto (mantenido por compatibilidad).\n"
        "Usar modulos individuales: wsaa, wsfev1, wsfexv1, wsbfev1, etc.")

    if len(sys.argv) == 1:
        sys.argv.append("/register")

    import win32com.server.register
    win32com.server.register.UseCommandLine(WSFE)
    win32com.server.register.UseCommandLine(WSBFE)
    win32com.server.register.UseCommandLine(WSFEX)
    win32com.server.register.UseCommandLine(WSCTG)
    win32com.server.register.UseCommandLine(wDigDepFiel)



#!/usr/bin/python
# -*- coding: utf-8 -*-
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTIBILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

"Módulo para Trazabilidad de Precursores Químicos RENPRE Resolución 900/12"

# Información adicional y documentación:
# http://www.sistemasagiles.com.ar/trac/wiki/TrazabilidadPrecursoresQuimicos

__author__ = "Mariano Reingart <reingart@gmail.com>"
__copyright__ = "Copyright (C) 2011 Mariano Reingart"
__license__ = "GPL 3.0+"
__version__ = "1.12a"

#http://renpre.servicios.pami.org.ar/portal_traza_renpre/paso5.html

import os
import socket
import sys
import datetime, time
import py3simplesoap.client
from py3simplesoap.client import SoapFault
from .utils import BaseWS, inicializar_y_capturar_excepciones, get_install_dir

HOMO = False
TYPELIB = False

WSDL = "https://servicios.pami.org.ar/trazamed.WebServiceSDRN?wsdl"
LOCATION = "https://servicios.pami.org.ar/trazamed.WebServiceSDRN?wsdl"
##WSDL = "https://trazabilidad.pami.org.ar:59050/trazamed.WebServiceSDRN?wsdl"  # prod.

class TrazaRenpre(BaseWS):
    "Interfaz para el WebService de Trazabilidad de Precursores Quimicos SEDRONAR SNT"
    _public_methods_ = ['SaveTransacciones',
                        'SendCancelacTransacc', 'GetTransaccionesWS',
                        'Conectar', 'LeerError', 'LeerTransaccion',
                        'SetUsername',
                        'SetParametro', 'GetParametro',
                        'GetCodigoTransaccion', 'GetResultado', 'LoadTestXML']

    _public_attrs_ = [
        'Username', 'Password',
        'CodigoTransaccion', 'Errores', 'Resultado',
        'XmlRequest', 'XmlResponse',
        'Version', 'InstallDir',
        'Traceback', 'Excepcion', 'LanzarExcepciones',
        ]

    _reg_progid_ = "TrazaRenpre"
    _reg_clsid_ = "{461298DB-0531-47CA-B3D9-B36FE6967209}"

    # Variables globales para BaseWS:
    HOMO = HOMO
    WSDL = WSDL
    Version = "%s %s %s" % (__version__, HOMO and 'Homologación' or '',
                            py3simplesoap.client.__version__)

    def __init__(self, reintentos=1):
        self.Username = self.Password = None
        BaseWS.__init__(self, reintentos)

    def inicializar(self):
        BaseWS.inicializar(self)
        self.CodigoTransaccion = self.Errores = self.Resultado = None

    def __analizar_errores(self, ret):
        "Comprueba y extrae errores si existen en la respuesta XML"
        self.Errores = ["%s: %s" % (it.get('_c_error', ""), it.get('_d_error', ""))
                        for it in ret.get('errores', [])]
        self.Resultado = ret.get('resultado')

    def Conectar(self, cache=None, wsdl=None, proxy="", wrapper=None, cacert=None, timeout=30):
        # Conecto usando el método estandard:
        ok = BaseWS.Conectar(self, cache, wsdl, proxy, wrapper, cacert, timeout,
                             soap_server="jetty")
        if ok:
            # si el archivo es local, asumo que ya esta corregido:
            if not self.wsdl.startswith("file"):
                # corrijo ubicación del servidor (localhost:9050 en el WSDL)
                location = self.wsdl[:-5]
                if 'IWebServiceSDRNService' in self.client.services:
                    ws = self.client.services['IWebServiceSDRNService']
                else:
                    ws = self.client.services['IWebServiceSDRN']
                ws['ports']['IWebServiceSDRNPort']['location'] = location
            # Establecer credenciales de seguridad:
            self.client['wsse:Security'] = {
                'wsse:UsernameToken': {
                    'wsse:Username': self.Username,
                    'wsse:Password': self.Password,
                    }
                }
        return ok

    @inicializar_y_capturar_excepciones
    def SaveTransacciones(self, usuario, password,
                         gln_origen=None, gln_destino=None, f_operacion=None,
                         id_evento=None, cod_producto=None, n_cantidad=None,
                         n_documento_operacion=None, n_remito=None,
                         id_tipo_transporte=None,
                         id_paso_frontera_ingreso=None,
                         id_paso_frontera_egreso=None,
                         id_tipo_documento_operacion=None,
                         d_dominio_tractor=None,
                         d_dominio_semi=None,
                         n_serie=None, n_lote=None, doc_despacho_plaza=None,
                         djai=None, n_cert_impo_expo=None,
                         id_tipo_documento=None, n_documento=None,
                         m_calidad_analitica=None, m_entrega_parcial=None,
                         doc_permiso_embarque=None, gln_transportista=None,
                         operacion_excento_djai=None, control_duplicidad=None,
                         ):
        "Permite informe por parte de un agente de una o varias transacciones"
        # creo los parámetros para esta llamada
        params = {  'gln_origen': gln_origen, 'gln_destino': gln_destino,
                    'f_operacion': f_operacion, 'id_evento': id_evento,
                    'cod_producto': cod_producto, 'n_cantidad': n_cantidad,
                    'n_documento_operacion': n_documento_operacion,
                    'n_remito': n_remito,
                    'id_tipo_transporte': id_tipo_transporte,
                    'id_paso_frontera_ingreso': id_paso_frontera_ingreso,
                    'id_paso_frontera_egreso': id_paso_frontera_egreso,
                    'id_tipo_documento_operacion': id_tipo_documento_operacion,
                    'd_dominio_tractor': d_dominio_tractor,
                    'd_dominio_semi': d_dominio_semi,
                    'n_serie': n_serie, 'n_lote': n_lote,
                    'doc_despacho_plaza': doc_despacho_plaza,
                    'djai': djai, 'n_cert_impo_expo': n_cert_impo_expo,
                    'id_tipo_documento': id_tipo_documento,
                    'n_documento': n_documento,
                    'm_calidad_analitica': m_calidad_analitica,
                    'm_entrega_parcial': m_entrega_parcial,
                    'doc_permiso_embarque': doc_permiso_embarque,
                    'gln_transportista': gln_transportista,
                    'operacion_excento_djai': operacion_excento_djai,
                    'control_duplicidad': control_duplicidad,
                    }
        # actualizo con parámetros generales:
        params.update(self.params_in)
        res = self.client.saveTransacciones(
            arg0=params,
            arg1=usuario,
            arg2=password,
        )
        ret = res['return']
        self.CodigoTransaccion = ret.get('codigoTransaccion')
        self.__analizar_errores(ret)
        return True

    @inicializar_y_capturar_excepciones
    def SendCancelacTransacc(self, usuario, password, codigo_transaccion):
        " Realiza la cancelación de una transacción"
        res = self.client.sendCancelaTransac(
            arg0=codigo_transaccion,
            arg1=usuario,
            arg2=password,
        )

        ret = res['return']

        self.CodigoTransaccion = ret['codigoTransaccion']
        self.__analizar_errores(ret)

        return True

    @inicializar_y_capturar_excepciones
    def SendConfirmaTransacc(self, usuario, password, p_ids_transac, f_operacion):
        "Confirma la recepción de un medicamento"
        res = self.client.sendConfirmaTransacc(
            arg0=usuario,
            arg1=password,
            arg2={'p_ids_transac': p_ids_transac, 'f_operacion': f_operacion},
        )
        ret = res['return']
        self.CodigoTransaccion = ret.get('id_transac_asociada')
        self.__analizar_errores(ret)
        return True

    @inicializar_y_capturar_excepciones
    def SendAlertaTransacc(self, usuario, password, p_ids_transac_ws):
        "Alerta un medicamento, acción contraria a “confirmar la transacción”."
        res = self.client.sendAlertaTransacc(
            arg0=usuario,
            arg1=password,
            arg2=p_ids_transac_ws,
        )
        ret = res['return']
        self.CodigoTransaccion = ret.get('id_transac_asociada')
        self.__analizar_errores(ret)
        return True

    @inicializar_y_capturar_excepciones
    def GetTransaccionesWS(self, usuario, password,
                p_id_transaccion_global=None,
                id_agente_origen=None, id_agente_destino=None,
                id_agente_informador=None,
                gtin=None, id_evento=None, cant_analitica=None,
                fecha_desde_op=None, fecha_hasta_op=None,
                fecha_desde_t=None, fecha_hasta_t=None,
                id_tipo=None,
                id_estado=None, nro_pag=1, cant_reg=100,
                ):
        "Obtiene los movimientos realizados y permite filtros de búsqueda"

        # preparo los parametros de entrada opcionales:
        kwargs = {}
        if p_id_transaccion_global is not None:
            kwargs['arg2'] = p_id_transaccion_global
        if id_agente_origen is not None:
            kwargs['arg3'] = id_agente_origen
        if id_agente_destino is not None:
            kwargs['arg4'] = id_agente_destino
        if id_agente_destino is not None:
            kwargs['arg5'] = id_agente_informador
        if gtin is not None:
            kwargs['arg6'] = gtin
        if id_evento is not None:
            kwargs['arg7'] = id_evento
        if cant_analitica is not None:
            kwargs['arg8'] = cant_analitica
        if fecha_desde_op is not None:
            kwargs['arg9'] = fecha_desde_op
        if fecha_hasta_op is not None:
            kwargs['arg10'] = fecha_hasta_op
        if fecha_desde_t is not None:
            kwargs['arg11'] = fecha_desde_t
        if fecha_hasta_t is not None:
            kwargs['arg12'] = fecha_hasta_t
        if id_tipo is not None:
            kwargs['arg13'] = id_tipo
        if id_estado is not None:
            kwargs['arg14'] = id_estado
        if nro_pag is not None:
            kwargs['arg15'] = nro_pag
        if cant_reg is not None:
            kwargs['arg16'] = cant_reg

        # llamo al webservice
        res = self.client.getTransaccionesWs(
            arg0=usuario,
            arg1=password,
            **kwargs
        )
        ret = res['return']
        if ret:
            self.__analizar_errores(ret)
            self.CantPaginas = ret.get('cantPaginas')
            self.HayError = ret.get('hay_error')
            self.TransaccionPlainWS = [it for it in ret.get('list', [])]
        return True

    def SetUsername(self, username):
        "Establezco el nombre de usuario"
        self.Username = username

    def SetPassword(self, password):
        "Establezco la contraseña"
        self.Password = password

    def GetCodigoTransaccion(self):
        "Devuelvo el código de transacción"
        return self.CodigoTransaccion

    def GetResultado(self):
        "Devuelvo el resultado"
        return self.Resultado



def main():
    "Función principal de pruebas (transaccionar!)"
    import os, time, sys
    global WSDL, LOCATION

    DEBUG = '--debug' in sys.argv

    ws = TrazaRenpre()

    ws.Username = 'testwservice'
    ws.Password = 'testwservicepsw'

    if '--prod' in sys.argv and not HOMO:
        WSDL = "https://trazabilidad.pami.org.ar:59050/trazamed.WebServiceSDRN?wsdl"
        print("Usando WSDL:", WSDL)
        sys.argv.pop(0)

    ws.Conectar("", WSDL)

    if ws.Excepcion:
        print(ws.Excepcion)
        print(ws.Traceback)
        sys.exit(-1)

    # print(ws.client.services)
    # op = ws.client.get_operation("sendMedicamentos")
    # import pdb;pdb.set_trace()
    if '--test' in sys.argv:
        ws.SaveTransacciones(
            usuario='pruebasws', password='pruebasws',
            gln_origen=8888888888888,
            gln_destino=8888888888888,
            f_operacion="20/05/2014",
            id_evento=44,
            cod_producto=88800000000035, # acido sulfúrico
            n_cantidad=1,
            n_documento_operacion=1,
            #m_entrega_parcial="",
            n_remito=123,
            n_serie=112,
            )
        print("Resultado", ws.Resultado)
        print("CodigoTransaccion", ws.CodigoTransaccion)
        print("Excepciones", ws.Excepcion)
        print("Erroes", ws.Errores)
    elif '--cancela' in sys.argv:
        ws.SendCancelacTransacc(*sys.argv[sys.argv.index("--cancela")+1:])
    elif '--consulta' in sys.argv:
        if '--movimientos' in sys.argv:
            ws.GetTransaccionesWS(
                                *sys.argv[sys.argv.index("--movimientos")+1:]
                                )
    else:
        ws.SaveTransacciones(*sys.argv[1:])
    print("|Resultado %5s|CodigoTransaccion %10s|Errores|%s|" % (
            ws.Resultado,
            ws.CodigoTransaccion,
            '|'.join(ws.Errores),
            ))
    if ws.Excepcion:
        print(ws.Traceback)

# busco el directorio de instalación (global para que no cambie si usan otra dll)
INSTALL_DIR = TrazaRenpre.InstallDir = get_install_dir()


if __name__ == '__main__':

    # ajusto el encoding por defecto (si se redirije la salida)
    if sys.stdout.encoding is None:
        import codecs, locale
        sys.stdout = codecs.getwriter(locale.getpreferredencoding())(sys.stdout,"replace");
        sys.stderr = codecs.getwriter(locale.getpreferredencoding())(sys.stderr,"replace");

    if '--register' in sys.argv or '--unregister' in sys.argv:
        import pythoncom
        import win32com.server.register
        win32com.server.register.UseCommandLine(TrazaRenpre)
    elif "/Automate" in sys.argv:
        # MS seems to like /automate to run the class factories.
        import win32com.server.localserver
        #win32com.server.localserver.main()
        # start the server.
        win32com.server.localserver.serve([TrazaRenpre._reg_clsid_])
    else:
        main()
